#! /usr/bin/python3
#
# dput/dput.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" dput — Debian package upload tool. """

import configparser
import email.parser
from hashlib import (
        md5,
        sha1,
        )
import importlib
import os
import os.path
import pkgutil
import re
import signal
import stat
import subprocess
import sys
import textwrap
import types

from . import configfile
from . import crypto
from .helper import dputhelper


app_library_path = os.path.dirname(__file__)

debug = 0


def import_upload_functions():
    """ Import upload method modules and make them available. """
    upload_methods = {}

    package_name = "methods"
    modules_path = os.path.join(app_library_path, package_name)
    modules_found = [
            name for (__, name, ispkg) in
            pkgutil.iter_modules([modules_path])
            if not ispkg]
    if debug:
        sys.stdout.write("D: modules_found: {!r}\n".format(modules_found))
    for module_name in modules_found:
        module = importlib.import_module("{package}.{module}".format(
                package=".".join(["dput", package_name]),
                module=module_name))
        if debug:
            sys.stdout.write("D: Module: {module_name} ({module!r})\n".format(
                    **vars()))
        method_name = module_name
        if debug:
            sys.stdout.write("D: Method name: {}\n".format(method_name))

        upload_methods[method_name] = module.upload

    return upload_methods


def parse_changes(changes_file):
    """ Parse the changes file. """
    check = changes_file.read(5)
    if check != "-----":
        changes_file.seek(0)
    else:
        # found a PGP header, gonna ditch the next 3 lines
        # eat the rest of the line
        changes_file.readline()
        # Hash: SHA1
        changes_file.readline()
        # empty line
        changes_file.readline()
    if not changes_file.readline().find("Format") != -1:
        changes_file.readline()
    changes_text = changes_file.read()
    changes = email.parser.HeaderParser().parsestr(changes_text)
    if 'files' not in changes:
        raise KeyError("No Files field in upload control file")
    for file_spec in changes['files'].strip().split("\n"):
        if len(file_spec.split()) != 5:
            sys.stderr.write(textwrap.dedent("""\
                    Invalid Files line in .changes:
                      {}
                    """).format(file_spec))
            sys.exit(1)
    return changes


def read_configs(config_files, debug):
    """ Read configuration settings from config files.

        :param config_files: Sequence of configuration files, open for reading.
        :param debug: If true, enable debugging output.
        :return: The resulting `ConfigParser` instance.

        The config parser will parse each file in `config_files`. Configuration
        from later files overrides earlier files.
        """
    config = configparser.ConfigParser()

    config.set('DEFAULT', 'login', "username")
    config.set('DEFAULT', 'method', "scp")
    config.set('DEFAULT', 'hash', "md5")
    config.set('DEFAULT', 'allow_unsigned_uploads', "0")
    config.set('DEFAULT', 'allow_dcut', "0")
    config.set('DEFAULT', 'distributions', "")
    config.set('DEFAULT', 'allowed_distributions', "")
    config.set('DEFAULT', 'run_lintian', "0")
    config.set('DEFAULT', 'run_dinstall', "0")
    config.set('DEFAULT', 'check_version', "0")
    config.set('DEFAULT', 'scp_compress', "0")
    config.set('DEFAULT', 'default_host_main', "")
    config.set('DEFAULT', 'post_upload_command', "")
    config.set('DEFAULT', 'pre_upload_command', "")
    config.set('DEFAULT', 'ssh_config_options', "")
    config.set('DEFAULT', 'passive_ftp', "1")
    config.set('DEFAULT', 'progress_indicator', "0")
    config.set('DEFAULT', 'delayed', "")

    for config_file in config_files:
        if debug:
            sys.stdout.write(
                    "D: Parsing configuration file ‘{path}’\n".format(
                        path=config_file.name))
        try:
            config.read_file(config_file)
        except configparser.ParsingError as e:
            sys.stderr.write(textwrap.dedent("""\
                    Error parsing configuration file:
                    {0}
                    """).format(str(e)))
            sys.exit(1)
        config_file.close()

    # only check for fqdn and incoming dir, rest have reasonable defaults
    error = 0
    for section in config.sections():
        if config.get(section, 'method') == "local":
            config.set(section, 'fqdn', "localhost")
        if (
                not config.has_option(section, 'fqdn') and
                config.get(section, 'method') != "local"):
            sys.stderr.write(
                    "Config error: {} must have a fqdn set\n".format(section))
            error = 1
        if not config.has_option(section, 'incoming'):
            sys.stderr.write((
                    "Config error: {} must have an incoming directory set\n"
                    ).format(section))
            error = 1
    if error:
        sys.exit(1)

    return config


def checksum_test(filename, hash_name):
    """ Get the hex string for the hash of a file's content.

        :param filename: Path to the file to read.
        :param hash_name: Name of the hash to use.
        :return: The computed hash value, as hexadecimal text.

        Currently supports md5, sha1. ripemd may come in the future.
        """
    try:
        file_to_test = open(filename, 'rb')
    except IOError:
        sys.stdout.write("Can't open {}\n".format(filename))
        sys.exit(1)

    if hash_name == 'md5':
        hash_type = md5
    else:
        hash_type = sha1

    check_obj = hash_type()

    while 1:
        data = file_to_test.read(65536)
        if len(data) == 0:
            break
        check_obj.update(data)

    file_to_test.close()
    checksum_text = check_obj.hexdigest()

    return checksum_text


def check_upload_variant(changes, debug):
    """ Check if this is a binary_upload only or not. """
    binary_upload = 0
    if 'architecture' in changes:
        arch = changes['architecture']
        if debug:
            sys.stdout.write("D: Architecture: {}\n".format(arch))
        if arch.find('source') < 0:
            if debug:
                sys.stdout.write("D: Doing a binary upload only.\n")
            binary_upload = 1
    return binary_upload


def verify_signature(
        host, changes_file_path, dsc_file_path,
        config, check_only, allow_unsigned_uploads, binary_upload,
        debug):
    """ Check the signature on the two files given via function call.

        :param host: Configuration host name.
        :param changes_file_path: Filesystem path of upload control file.
        :param dsc_file_path: Filesystem path of source control file.
        :param config: `ConfigParser` instance for this application.
        :param check_only: If true, no upload is requested.
        :param allow_unsigned_uploads: If true, allow an unsigned upload.
        :param binary_upload: If true, this upload excludes source.
        :param debug: If true, enable debugging output.
        :return: ``None``.
        """

    def assert_good_signature_or_exit(path):
        """ Assert the signature on the file at `path` is good. """
        try:
            with open(path, encoding='utf-8') as infile:
                crypto.check_file_signature(infile)
        except Exception as exc:
            if isinstance(exc, crypto.gpg.errors.GPGMEError):
                sys.stdout.write("{}\n".format(exc))
                sys.exit(1)
            else:
                raise

    if debug:
        sys.stdout.write(
                "D: upload control file: {}\n".format(changes_file_path))
        sys.stdout.write(
                "D: source control file: {}\n".format(dsc_file_path))
    if ((check_only or config.getboolean(host, 'allow_unsigned_uploads') == 0)
            and not allow_unsigned_uploads):
        sys.stdout.write("Checking signature on .changes\n")
        assert_good_signature_or_exit(changes_file_path)
        if not binary_upload:
            sys.stdout.write("Checking signature on .dsc\n")
            assert_good_signature_or_exit(dsc_file_path)


# Regular expression to match the components of a Debian package version
# string.
# Documentation: Debian Policy §5.6.12. “Version”:
#     The format is: “[epoch:]upstream_version[-debian_revision]”.
version_regex = re.compile(
        "^"
        # Epoch is an unsigned decimal integer.
        "((?P<epoch>[0-9]+):)?"
        # Upstream version is either followed by a hyphen (implying a Debian
        # revision), or there are no hyphens (implying a native package).
        "(?P<upstream_version>(?:[0-9a-zA-Z.~+-]+(?=-)|[0-9a-zA-Z.~+]+))"
        # The Debian revision is separated by the last hyphen in the string.
        "(-(?P<debian_version>[0-9a-zA-Z.~+]+))?"
        "$")


def source_check(changes, debug):
    """ Check if a source tarball has to be included in the package or not. """
    include_orig = include_tar = 0
    if 'version' in changes:
        version = changes['version']
        if debug:
            sys.stdout.write("D: Package Version: {}\n".format(version))
        version_parts = version_regex.match(version)
        if version_parts is None:
            raise dputhelper.UploadChangesInvalidVersionDputError(version)
        (epoch, upstream_version, debian_version) = (
                version_parts.group('epoch'),
                version_parts.group('upstream_version'),
                version_parts.group('debian_version'))
        if (debian_version is None):
            # The version string implies this is a Debian native package.
            include_tar = 1
        else:
            if (epoch is not None):
                if debug:
                    sys.stdout.write("D: Epoch found\n")
            if debug:
                sys.stdout.write(
                        "D: Upstream Version: {}\n".format(upstream_version))
                sys.stdout.write(
                        "D: Debian Version: {}\n".format(debian_version))
            if debian_version in ["0.1", "1", "1.1", "0ubuntu1"]:
                include_orig = 1
            else:
                include_tar = 1
    return (include_orig, include_tar)


def verify_files(
        changes_file_directory, changes_file_name, host,
        config, check_only, check_version, allow_unsigned_uploads,
        debug):
    """ Run some tests on the files to verify that they are in good shape.

        :param changes_file_directory: Directory path of the upload
            control file.
        :param changes_file_name: Filename of the upload control file.
        :param host: Configuration host name.
        :param config: `ConfigParser` instance for this application.
        :param check_only: If true, no upload is requested.
        :param check_version: If true, check the package version
            before upload.
        :param allow_unsigned_uploads: If true, allow an unsigned upload.
        :param debug: If true, enable debugging output.
        :return: A collection of filesystem paths of all files to upload.
        """
    include_orig_tar_gz = include_tar_gz = 0
    files_to_upload = []

    changes_file_path = os.path.join(changes_file_directory, changes_file_name)

    if debug:
        sys.stdout.write(
                "D: Validating contents of changes file {}\n".format(
                    changes_file_path))
    try:
        changes_file = open(changes_file_path, 'r', encoding='UTF-8')
    except IOError:
        sys.stdout.write("Can't open {}\n".format(changes_file_path))
        sys.exit(1)
    changes = parse_changes(changes_file)
    changes_file.close

    # Find out if it's a binary only upload or not
    binary_upload = check_upload_variant(changes, debug)

    if binary_upload:
        dsc_file_path = ""
    else:
        dsc_file_path = None
        for file_spec in changes['files'].strip().split("\n"):
            # Filename only.
            file_name = file_spec.split()[4]
            if file_name.find(".dsc") != -1:
                if debug:
                    sys.stdout.write("D: dsc-File: {}\n".format(file_name))
                dsc_file_path = os.path.join(changes_file_directory, file_name)
        if not dsc_file_path:
            sys.stderr.write("Error: no dsc file found in sourceful upload\n")
            sys.exit(1)

    # Run the check to verify that the package has been tested.
    try:
        if config.getboolean(host, 'check_version') == 1 or check_version:
            version_check(changes_file_directory, changes, debug)
    except configparser.NoSectionError as e:
        sys.stderr.write("Error in config file:\n{}\n".format(e))
        sys.exit(1)

    # Verify the signature of the maintainer
    verify_signature(
            host, changes_file_path, dsc_file_path,
            config, check_only, allow_unsigned_uploads, binary_upload,
            debug)

    # Check the sources
    (include_orig_tar_gz, include_tar_gz) = source_check(changes, debug)

    # Check md5sum and the size
    file_list = changes['files'].strip().split("\n")
    hash_name = config.get('DEFAULT', 'hash')
    for file_spec in file_list:
        (check_sum, size, section, priority, file_name) = file_spec.split()
        file_path = os.path.join(changes_file_directory, file_name)
        if debug:
            sys.stdout.write("D: File to upload: {}\n".format(file_path))
        if checksum_test(file_path, hash_name) != check_sum:
            if debug:
                sys.stdout.write(
                        "D: Checksum from .changes: {}\n".format(check_sum))
                sys.stdout.write(
                        "D: Generated Checksum: {}\n".format(
                            checksum_test(file_path, hash_name)))
            sys.stdout.write(
                    "Checksum doesn't match for {}\n".format(file_path))
            sys.exit(1)
        else:
            if debug:
                sys.stdout.write(
                        "D: Checksum for {} is fine\n".format(file_path))
        if os.stat(file_path)[stat.ST_SIZE] != int(size):
            if debug:
                sys.stdout.write("D: size from .changes: {}\n".format(size))
                sys.stdout.write(
                        "D: calculated size: {}\n".format(
                            os.stat(file_path)[stat.ST_SIZE]))
            sys.stdout.write(
                    "size doesn't match for {}\n".format(file_path))

        files_to_upload.append(file_path)

    # Check filenames.
    for file_path in files_to_upload:
        if (file_path.endswith(".orig.tar.gz") and not include_orig_tar_gz):
            if debug:
                sys.stdout.write("D: Filename: {}\n".format(file_path))
                sys.stdout.write("D: Suffix: {}\n\n".format(file_path[-12:]))
            sys.stdout.write(
                    "Package includes an .orig.tar.gz file although"
                    " the debian revision suggests\n"
                    "that it might not be required.\n")
        elif (
                file_path.endswith(".tar.gz") and not include_tar_gz
                and not include_orig_tar_gz):
            if debug:
                sys.stdout.write("D: Filename: {}\n".format(file_path))
                sys.stdout.write("D: Suffix: {}\n".format(file_path[-7:]))
            sys.stdout.write(
                    "Package includes a .tar.gz file although"
                    " the version suggests that it might\n"
                    "not be required.\n")

    distribution = changes.get('distribution')
    allowed_distributions = config.get(host, 'allowed_distributions')
    if distribution and allowed_distributions:
        if debug:
            sys.stdout.write((
                    "D: Checking: distribution {distribution}"
                    " matches {allowed_distributions}\n"
                    ).format(**vars()))
        if not re.match(allowed_distributions, distribution):
            raise dputhelper.DputUploadFatalException((
                    "Error: uploading files"
                    " for distribution {distribution} to {host} not allowed."
                    ).format(**vars()))

    if debug:
        sys.stdout.write("D: File to upload: {}\n".format(changes_file_path))
    files_to_upload.append(changes_file_path)

    return files_to_upload


def print_config(config, debug):
    """ Print the configuration and exit. """
    sys.stdout.write("\n")
    config.write(sys.stdout)
    sys.stdout.write("\n")


def print_default_upload_method(config):
    """ Print the default upload method defined in the configuration.

        :param config: `ConfigParser` instance for this application.
        :return: ``None``.
        """
    sys.stdout.write(textwrap.dedent("""\

            Default Method: {}

            """.format(config.get('DEFAULT', 'method'))))


def print_host_list(config):
    """ Print a list of hosts defined in the configuration.

        :param config: `ConfigParser` instance for this application.
        :return: ``None``.
        """
    for section in config.sections():
        distributions = ""
        if config.get(section, 'distributions'):
            distributions = ", distributions: {}".format(
                    config.get(section, 'distributions'))
        sys.stdout.write((
                "{section} => {fqdn}"
                "  (Upload method: {method}{distributions})\n"
                ).format(
                    section=section,
                    fqdn=config.get(section, 'fqdn'),
                    method=config.get(section, 'method'),
                    distributions=distributions))
    sys.stdout.write("\n")


def create_upload_file(
        changes_file_name, host, fqdn,
        changes_file_directory, files_to_upload,
        debug):
    """ Write the log file for the upload.

        :param changes_file_name: File name of package to upload.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of the remote host.
        :param changes_file_directory: Directory path of the upload
            control file.
        :param debug: If true, enable debugging output.
        :return: ``None``.

        The upload log file is named ‘basename.hostname.upload’, where
        “basename” is the package file name without suffix, and
        “hostname” is the name of the host as specified in the
        configuration file.

        For example, uploading ‘foo_1.2.3-1_xyz.deb’ to host ‘bar’
        will be logged to ‘foo_1.2.3-1_xyz.bar.upload’.

        The upload log file is written to the
        directory containing the upload control file.
        """
    # only need first part
    changes_file_name_base = os.path.splitext(changes_file_name)[0]
    log_file_name = "{name_stem}.{host}.upload".format(
            name_stem=changes_file_name_base,
            **vars())
    log_file_path = os.path.join(changes_file_directory, log_file_name)
    if debug:
        sys.stdout.write("D: Writing logfile: {}\n".format(log_file_path))
    try:
        if os.access(log_file_path, os.R_OK):
            log_file = open(log_file_path, 'a', encoding='UTF-8')
        else:
            log_file = open(log_file_path, 'w', encoding='UTF-8')
    except IOError:
        sys.stderr.write("Could not write {}\n".format(log_file_path))
        sys.exit(1)

    for file_path in files_to_upload:
        log_entry = (
                "Successfully uploaded {file_name} to {fqdn} for {host}.\n"
                ).format(
                        file_name=os.path.basename(file_path),
                        **vars())
        log_file.write(log_entry)
    log_file.close()


def run_lintian_test(changes_file):
    """ Run lintian on the changes file and stop if it finds errors. """

    if os.access(changes_file, os.R_OK):
        if os.access("/usr/bin/lintian", os.R_OK):
            old_signal = signal.signal(signal.SIGPIPE, signal.SIG_DFL)
            sys.stdout.write("Package is now being checked with lintian.\n")
            if dputhelper.check_call(
                    ["lintian", "-i", changes_file]
                    ) != dputhelper.EXIT_STATUS_SUCCESS:
                sys.stdout.write(
                        "\n"
                        "Lintian says this package is not compliant"
                        " with the current policy.\n"
                        "Please check the current policy and your package.\n"
                        "Also see lintian documentation about overrides.\n")
                sys.exit(1)
            else:
                signal.signal(signal.SIGPIPE, old_signal)
                return 0
        else:
            sys.stdout.write(
                    "lintian is not installed, skipping package test.\n")
    else:
        sys.stdout.write("Can't read {}\n".format(changes_file))
        sys.exit(1)


def guess_upload_host(changes_file_directory, changes_file_name, config):
    """ Guess the host where the package should be uploaded to.

        :param changes_file_directory: Directory path of the upload
            control file.
        :param changes_file_name: Filename of the upload control file.
        :param config: `ConfigParser` instance for this application.
        :return: The hostname determined for this upload.

        This is based on information from the upload control
        (‘*.changes’) file.
        """
    distribution = ""
    dist_re = re.compile(r'^Distribution: (.*)')

    changes_file_path = os.path.join(changes_file_directory, changes_file_name)

    try:
        changes_file = open(changes_file_path, 'r', encoding='UTF-8')
    except IOError:
        sys.stdout.write("Can't open {}\n".format(changes_file_path))
        sys.exit(1)
    lines = changes_file.readlines()
    for line in lines:
        match = dist_re.search(line)
        if match:
            distribution = match.group(1)

    # Try to guess a host based on the Distribution: field
    if distribution:
        for section in config.sections():
            host_dists = config.get(section, 'distributions')
            if not host_dists:
                continue
            for host_dist in host_dists.split(","):
                if distribution == host_dist.strip():
                    if debug:
                        sys.stdout.write((
                                "D: guessing host {section}"
                                " based on distribution {host_dist}\n"
                                ).format(**vars()))
                    return section

    if len(config.get('DEFAULT', 'default_host_main')) != 0:
        sys.stdout.write(
                "Trying to upload package to {}\n".format(
                    config.get('DEFAULT', 'default_host_main')))
        return config.get('DEFAULT', 'default_host_main')
    else:
        sys.stdout.write(
                "Trying to upload package to ftp-master"
                " (ftp.upload.debian.org)\n")
        return "ftp-master"


def get_upload_method_name_for_host(config, host):
    """ Get the upload method name from the config for the host.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :return: The name of the upload method for the specified host.
        """
    method_name = None
    if not config.get(host, 'method'):
        method_name = config.get('DEFAULT', 'method')
    else:
        method_name = config.get(host, 'method')

    return method_name


def get_login_for_host(config, host, debug=False):
    """ Get the login username from the config for the host.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :param debug: If true, enable debugging output.
        :return: The username for the specified host, if it is
            configured; otherwise, the system username.
        """
    login = None
    if (
            config.has_option(host, 'login')
            and config.get(host, 'login') != 'username'):
        login = config.get(host, 'login')
        if debug:
            sys.stdout.write(
                    "D: Login {login} from section {host} used\n".format(
                        **vars()))
    elif (
            config.has_option('DEFAULT', 'login')
            and config.get('DEFAULT', 'login') != 'username'):
        login = config.get('DEFAULT', 'login')
        if debug:
            sys.stdout.write("D: Default login {login} used\n".format(
                **vars()))
    else:
        login = dputhelper.get_username_from_system(debug=debug)
        sys.stdout.write((
                "D: Neither host {host} nor default login used."
                " Using {login}\n"
                ).format(**vars()))

    return login


def get_fqdn_for_host(config, host):
    """ Get the FQDN to use, determined by the `host` config section.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :return: The fully qualified domain name (FQDN) for the
            specified host.

        As a special case, if the host configuration's 'method' option
        specifies the value "local", the corresponding FQDN is always
        "localhost". Otherwise, the FQDN value is the host
        configuration's 'fqdn' option value.
        """
    if config.get(host, 'method') == "local":
        fqdn = "localhost"
    else:
        config_fqdn = config.get(host, 'fqdn')
        if ":" in config_fqdn:
            # The value includes a port specification we don't want.
            (fqdn, __) = config_fqdn.split(":", 1)
        else:
            fqdn = config_fqdn

    return fqdn


def get_port_for_host(config, host):
    """ Get the port to use, determined by the `host` config section.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :return: The TCP port if specified, otherwise ``None``.

        If the host config specified a 'fqdn' value of the form
        `"example.com:port"`, the `port` value is returned. Otherwise,
        ``None``.
        """
    port_text = None
    if config.get(host, 'method') != "local":
        config_fqdn = config.get(host, 'fqdn')
        if ":" in config_fqdn:
            (__, port_text) = config_fqdn.split(":", 1)
    port = int(port_text) if port_text else None

    return port


def get_incoming_for_host(config, host):
    """ Get the incoming directory to use, determined by the host config.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :return: The incoming directory path for the remote host.
        """
    config_incoming = config.get(host, 'incoming')
    result = config_incoming.rstrip(os.path.sep)

    return result


def get_delayed_days(days_text):
    """ Get the number of days specified for delayed upload.

        :param days_text: The text specifying number of days to delay.
        :raise TypeError: When the `days_text` is not a valid integer
            representation.
        :raise ValueError: When the integer represented by `days_text` is
            outside the valid range {0..15}.
        :return: The number of days as an integer, or ``None`` if `days_text`
            is ``None``.
        """
    result = None

    try:
        days = int(days_text)
    except (TypeError, ValueError) as exc:
        raise TypeError(
                "delayed days value must be a decimal integer: "
                "{}".format(days_text)
                ) from exc
    if days not in range(16):
        raise ValueError(
                "delayed days value must be from 0 to 15: "
                "{:d}".format(days))
    result = days

    return result


def determine_delay_days(options, host_config):
    """ Determine delay days from `options` and `host_config` collections.

        :param options: The command-line options, as attributes in a namespace.
        :param host_config: The ConfigParser section to interrogate.
        :raise TypeError: When an invalid value is specified.
        :raise ValueError: When an out-of-range value is specified.
        :return: The integer delay days value, or ``None`` if not specified.

        For configuration options: Interpret empty string (``""``) as ``None``.
        """
    result = None
    commandline_text_value = (
            options.delay_days_text if hasattr(options, 'delay_days_text')
            else NotImplemented)
    config_text_value = host_config.get('delayed', fallback=NotImplemented)
    if config_text_value == "":
        config_text_value = None

    text_value = (
        commandline_text_value if commandline_text_value is not NotImplemented
        else config_text_value)
    if (
            text_value is not NotImplemented
            and text_value is not None
    ):
        result = get_delayed_days(text_value)

    return result


def get_progress_indicator_for_host(config, host):
    """ Get the code for the progress indicator to use for this upload. """
    result = config.getint(host, 'progress_indicator')
    if not os.isatty(1):
        result = 0
    return result


def verify_config_upload_methods(config, host, upload_methods, debug=False):
    """ Verify the configured upload method for host and default.

        :param config: The `ConfigParser` instance for this application.
        :param host: The remote host for which to get the configuration.
        :param upload_methods: The mapping from upload method name to
            function.
        :param debug: If true, enable debugging output.
        :return: ``None``.

        If the configured upload method for the host or for default do
        not verify, the program exits.
        """
    if debug:
        sys.stdout.write(
                "D: Default Method: {}\n".format(
                    config.get('DEFAULT', 'method')))
    if config.get('DEFAULT', 'method') not in upload_methods:
        sys.stdout.write(
                "Unknown upload method: {}\n".format(
                    config.get('DEFAULT', 'method')))
        sys.exit(1)
    if debug:
        sys.stdout.write(
                "D: Host Method: {}\n".format(config.get(host, 'method')))
    if config.get(host, 'method') not in upload_methods:
        sys.stdout.write(
                "Unknown upload method: {}\n".format(
                    config.get(host, 'method')))
        sys.exit(1)


def dinstall_caller(filename, host, fqdn, login, incoming, debug):
    """ Run ‘dinstall’ for the package on the remote host.

        :param filename: Debian package filename to install.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of the remote host.
        :param login: Username for login to the remote host.
        :param incoming: Filesystem path on remote host for incoming
            packages.
        :param debug: If true, enable debugging output.
        :return: ``None``.

        Run ‘dinstall’ on the remote host in test mode, and present
        the output to the user.

        This is so the user can see if the package would be installed
        or not.
        """
    command = [
            "ssh", "{login}@{fqdn}".format(**vars()),
            "cd", str(incoming),
            ";", "dinstall", "-n", str(filename)]
    if debug:
        sys.stdout.write(
                "D: Logging into {login}@{host}:{incoming}\n".format(**vars()))
        sys.stdout.write("D: dinstall -n {}\n".format(filename))
    if dputhelper.check_call(command) != dputhelper.EXIT_STATUS_SUCCESS:
        sys.stdout.write(
                "Error occured while trying to connect, or while"
                " attempting to run dinstall.\n")
        sys.exit(1)


def version_check(path, changes, debug):
    """ Check if the caller has installed the package also on his system.

        This is for testing purposes before uploading it. If not, we
        reject the upload.
        """
    packages_to_check = []

    # Get arch
    dpkg_proc = subprocess.Popen(
            "dpkg --print-architecture",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            shell=True, close_fds=True)
    dpkg_stdout = dputhelper.make_text_stream(dpkg_proc.stdout)
    dpkg_stderr = dputhelper.make_text_stream(dpkg_proc.stderr)
    dpkg_output = dpkg_stdout.read()
    dpkg_architecture = dpkg_output.strip()
    dpkg_stdout.close()
    dpkg_stderr_output = dpkg_stderr.read()
    dpkg_stderr.close()
    if debug and dpkg_stderr_output:
        sys.stdout.write((
                "D: dpkg-architecture stderr output:"
                " {!r}\n").format(dpkg_stderr_output))
    if debug:
        sys.stdout.write(
                "D: detected architecture: '{}'\n".format(dpkg_architecture))

    # Get filenames of deb files:
    for file_spec in changes['files'].strip().split("\n"):
        file_name = os.path.join(path, file_spec.split()[4])
        if file_name.endswith(".deb"):
            if debug:
                sys.stdout.write("D: Debian Package: {}\n".format(file_name))
            dpkg_proc = subprocess.Popen(
                    "dpkg --field {}".format(file_name),
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                    shell=True, close_fds=True)
            dpkg_stdout = dputhelper.make_text_stream(dpkg_proc.stdout)
            dpkg_stderr = dputhelper.make_text_stream(dpkg_proc.stderr)
            dpkg_output = dpkg_stdout.read()
            dpkg_stdout.close()
            dpkg_fields = email.parser.HeaderParser().parsestr(dpkg_output)
            dpkg_stderr_output = dpkg_stderr.read()
            dpkg_stderr.close()
            if debug and dpkg_stderr_output:
                sys.stdout.write((
                        "D: dpkg stderr output:"
                        " {!r}\n").format(dpkg_stderr_output))
            if (
                    dpkg_architecture
                    and dpkg_fields['architecture'] not in [
                        'all', dpkg_architecture]):
                if debug:
                    sys.stdout.write((
                            "D: not install-checking {} due to arch mismatch\n"
                            ).format(file_name))
            else:
                package_name = dpkg_fields['package']
                version_number = dpkg_fields['version']
                if debug:
                    sys.stdout.write(
                            "D: Package to Check: {}\n".format(package_name))
                if debug:
                    sys.stdout.write(
                            "D: Version to Check: {}\n".format(version_number))
                packages_to_check.append((package_name, version_number))

    for package_name, version_to_check in packages_to_check:
        if debug:
            sys.stdout.write("D: Name of Package: {}\n".format(package_name))
        dpkg_proc = subprocess.Popen(
                "dpkg -s {}".format(package_name),
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                shell=True, close_fds=True)
        dpkg_stdout = dputhelper.make_text_stream(dpkg_proc.stdout)
        dpkg_stderr = dputhelper.make_text_stream(dpkg_proc.stderr)
        dpkg_output = dpkg_stdout.read()
        dpkg_stdout.close()
        dpkg_fields = email.parser.HeaderParser().parsestr(dpkg_output)
        dpkg_stderr_output = dpkg_stderr.read()
        dpkg_stderr.close()
        if debug and dpkg_stderr_output:
            sys.stdout.write((
                    "D: dpkg stderr output:"
                    " {!r}\n").format(dpkg_stderr_output))
        if 'version' in dpkg_fields:
            installed_version = dpkg_fields['version']
            if debug:
                sys.stdout.write(
                        "D: Installed-Version: {}\n".format(installed_version))
            if debug:
                sys.stdout.write(
                        "D: Check-Version: {}\n".format(version_to_check))
            if installed_version != version_to_check:
                sys.stdout.write(
                        "Package to upload is not installed, but it appears"
                        " you have an older version installed.\n")
        else:
            sys.stdout.write(
                    "Uninstalled Package. Test it before uploading it.\n")
            sys.exit(1)


def execute_command(command, position, debug=False):
    """ Run `command` in a child shell.

        :param command: Command line to execute.
        :param position: Position of the command: 'pre' or 'post'.
        :param debug: If true, enable debugging output.
        :return: ``None``.
        """
    if debug:
        sys.stdout.write("D: Command: {}\n".format(command))
    if subprocess.call(command, shell=True):
        raise dputhelper.DputUploadFatalException(
                "Error: {} upload command failed.".format(position))


def check_upload_logfile(
        changes_file_path, host, fqdn,
        check_only, force_upload, debug):
    """ Check if the user already put this package on the specified host.

        :param changes_file_path: Filesystem path of upload control file.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of the remote host.
        :param check_only: If true, no upload is requested.
        :param force_upload: If true, don't check the upload log file.
        :param debug: If true, enable debugging output.
        :return: ``None``.
        """
    uploaded = 0
    upload_logfile_path = "{filename_stem}.{host}.upload".format(
            filename_stem=(os.path.splitext(changes_file_path)[0]),
            **vars())
    if not check_only and not force_upload:
        if not os.path.exists(upload_logfile_path):
            return
        try:
            upload_logfile = open(upload_logfile_path)
        except IOError:
            sys.stderr.write("Couldn't open {}\n".format(upload_logfile_path))
            sys.exit(1)
        for line in upload_logfile.readlines():
            if line.find(fqdn) != -1:
                uploaded = 1
        if uploaded:
            sys.stderr.write(textwrap.dedent("""\
                    Package has already been uploaded to {host} on {fqdn},
                    recorded in upload log file ‘{upload_logfile_path}’.
                    Not attempting upload.
                    """).format(**vars()))
            sys.stderr.write(textwrap.dedent("""\
                    To ignore this check and attempt upload anyway,
                    use the ‘--force’ option.
                    """))
            sys.stderr.write(
                    "Nothing more to do for {}\n".format(changes_file_path))
            sys.exit(os.EX_OK)


def make_usage_message():
    """ Make the program usage help message. """
    text = textwrap.dedent("""\
        Usage: dput [options] [host] <package(s).changes>
         Supported options (see man page for long forms):
           -c: Config file to parse.
           -d: Enable debug messages.
           -D: Run dinstall after upload.
           -e: Upload to a delayed queue. Takes an argument from 0 to 15.
           -f: Force an upload.
           -h: Display this help message.
           -H: Display a list of hosts from the config file.
           -l: Run lintian before upload.
           -U: Do not write a .upload file after uploading.
           -o: Only check the package.
           -p: Print the configuration.
           -P: Use passive mode for ftp uploads.
           -s: Simulate the upload only.
           -u: Don't check GnuPG signature.
           -v: Display version information.
           -V: Check the package version and then upload it.
        """)
    return text


def make_delayed_queue_path(days=None):
    """ Make the relative DELAYED queue path for the specified delay.

        :param days: Number of days (integer).
        :return: The full queue name, if `days` is a number;
            otherwise, the empty string.
        """
    queue_fullname = ""
    if days is not None:
        queue_name = "{days:d}-day".format(days=days)
        queue_fullname = "DELAYED/{queue}".format(queue=queue_name)
    return queue_fullname


def upload_files_via_simulate(
        method, config, host, fqdn, login, incoming, files_to_upload,
        debug=False):
    """ Simulate the upload of files.

        :param method: Name of requested upload method.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :return: ``None``.

        The upload will not actually be performed, but diagnostic messages
        will be emitted.
        """
    sys.stderr.write(
            "Not performing upload: ‘simulate’ option specified.\n")
    for file_path in files_to_upload:
        sys.stderr.write((
                "Uploading with {method}: {file_path}"
                " to {fqdn}:{incoming}\n"
                ).format(**vars()))


def upload_files_via_method_local(
        method_name, method, config, host, fqdn, login, incoming,
        files_to_upload, debug,
        compress=False, progress=0):
    """ Upload files via ‘local’ method.

        :param method_name: Name of requested upload method.
        :param method: Callable upload method to invoke.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :param compress: Legacy dummy parameter.
        :param progress: Code specifying progress indicator to use.
        :return: ``None``.
        """
    method(
            fqdn, login, incoming,
            files_to_upload, debug,
            compress, progress)


def upload_files_via_method_ftp(
        method_name, method, config, host, fqdn, login, incoming,
        files_to_upload, debug,
        commandline_passive=None, progress=0):
    """ Upload files via ‘ftp’ method.

        :param method_name: Name of requested upload method.
        :param method: Callable upload method to invoke.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :param commandline_passive: Boolean value specifying whether
            the command line option "--passive" was received.
        :param progress: Code specifying progress indicator to use.
        :return: ``None``.
        """
    port = get_port_for_host(config, host)
    if not port:
        port = 21
    ftp_passive = config.getboolean(host, 'passive_ftp')
    if commandline_passive is not None:
        ftp_passive = commandline_passive
    ftp_mode_name = ["active", "passive"][ftp_passive]
    if debug:
        sys.stdout.write("D: FTP port: {port:d}\n".format(port=port))
        sys.stdout.write("D: Using {mode} ftp\n".format(mode=ftp_mode_name))

    method(
            fqdn, login, incoming,
            files_to_upload, debug,
            ftp_mode=ftp_passive, progress=progress, port=port)


def upload_files_via_method_scp(
        method_name, method, config, host, fqdn, login, incoming,
        files_to_upload, debug,
        compress=False, ssh_config_options=None):
    """ Upload files via ‘scp’ method.

        :param method_name: Name of requested upload method.
        :param method: Callable upload method to invoke.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :param compress: Boolean value specifying whether to
            enable SSH compression.
        :param ssh_config_options: Sequence of SSH configuration
            options to pass on SSH command line; or ``None`` if no
            options.
        :return: ``None``.
        """
    if ssh_config_options is None:
        ssh_config_options = []
    if debug:
        if compress:
            sys.stdout.write("D: Setting compression for scp\n")
        sys.stdout.write("\n  ".join(
                ["D: ssh config options:", ""]
                + ssh_config_options
                + [""]))

    method(
            fqdn, login, incoming,
            files_to_upload, debug,
            compress=compress, ssh_config_options=ssh_config_options)


def upload_files_via_method_sftp(
        method_name, method, config, host, fqdn, login, incoming,
        files_to_upload, debug,
        compress=None, progress=0):
    """ Upload files via ‘sftp’ method.

        :param method_name: Name of requested upload method.
        :param method: Callable upload method to invoke.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :param compress: Boolean value specifying whether to
            enable SSH compression.
        :param progress: Code specifying progress indicator to use.
        :return: ``None``.
        """

    method(
            fqdn, login, incoming,
            files_to_upload, debug,
            compress=compress, progress=progress)


def upload_files_via_method(
        method_name, method, config, host, fqdn, login, incoming,
        files_to_upload, debug,
        dummy=0, progress=0):
    """ Upload files via specified method.

        :param method_name: Name of requested upload method.
        :param method: Callable upload method to invoke.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param fqdn: Fully-qualified domain name of remote host.
        :param login: Username to use for login to remote host.
        :param incoming: Directory on remote host for incoming files.
        :param files_to_upload: Collection of file names to upload.
        :param debug: If true, enable debugging output.
        :param dummy: Legacy dummy parameter.
        :param progress: Code specifying progress indicator to use.
        :return: ``None``.
        """
    method(
            fqdn, login, incoming,
            files_to_upload, debug,
            dummy=dummy, progress=progress)


def upload_files(
        upload_methods, config, host, files_to_upload,
        simulate=False, delay_days=None, commandline_passive=None,
        debug=False):
    """ Upload files to the host.

        :param upload_methods: Mapping of {method_name: callable}.
        :param config: `ConfigParser` instance for this application.
        :param host: Configuration host name.
        :param files_to_upload: Collection of file names to upload.
        :param simulate: If true, simulate the upload only.
        :param delay_days: Integer number of days for incoming delay queue.
        :param commandline_passive: If true, enable FTP passive mode.
        :param debug: If true, enable debugging output.
        :return: ``None``.

        Upload the specified files to the host, using the method and
        credentials from the configuration for the host.
        """
    method = config.get(host, 'method')
    fqdn = get_fqdn_for_host(config, host)
    login = get_login_for_host(config, host, debug=debug)
    incoming = get_incoming_for_host(config, host)
    delayed_queue = make_delayed_queue_path(days=delay_days)
    if delayed_queue:
        incoming = os.path.join(incoming, delayed_queue)
    progress = get_progress_indicator_for_host(config, host)

    if not simulate:
        destination_description = host
        if delayed_queue:
            destination_description += " [{queue}]".format(
                    queue=delayed_queue)
        sys.stdout.write((
                "Uploading to {destination}"
                " (via {method} to {fqdn}):\n"
                ).format(
                    destination=destination_description,
                    **vars()))
        if debug:
            sys.stdout.write("D: FQDN: {}\n".format(fqdn))
            sys.stdout.write("D: Login: {}\n".format(login))
            sys.stdout.write("D: Incoming: {}\n".format(incoming))
        if method == 'local':
            upload_func = upload_files_via_method_local
            upload_func_kwargs = {
                    'compress': 0,
                    'progress': progress,
                    }
        elif method == 'ftp':
            upload_func = upload_files_via_method_ftp
            upload_func_kwargs = {
                    'commandline_passive': commandline_passive,
                    'progress': progress,
                    }
        elif method == 'scp':
            scp_compress = config.getboolean(host, 'scp_compress')
            ssh_config_options = [
                    option_text for option_text in (
                        line.strip() for line in
                        config.get(host, 'ssh_config_options').split("\n"))
                    if option_text]
            upload_func = upload_files_via_method_scp
            upload_func_kwargs = {
                    'compress': scp_compress,
                    'ssh_config_options': ssh_config_options,
                    }
        elif method == 'sftp':
            upload_func = upload_files_via_method_sftp
            upload_func_kwargs = {'compress': 1}
        else:
            upload_func = upload_files_via_method
            upload_func_kwargs = {
                    'dummy': 0,
                    'progress': progress,
                    }

        upload_func(
                method, upload_methods[method], config, host,
                fqdn, login, incoming, files_to_upload, debug,
                **upload_func_kwargs)

    else:
        upload_files_via_simulate(
                method, config, host, fqdn, login, incoming,
                files_to_upload, debug=debug)


def parse_commandline(argv):
    """ Parse the command-line arguments to option values.

        :param argv: The command-line arguments, as a sequence of text
            strings.
        :return: A 2-tuple (`options`, `args`), where `options` is the
            collection of option values; `args` is the remaining arguments
            after removing all options.

        The `options` collection is a `SimpleNamespace` instance, with
        each value represented as a named attribute.
        """
    global debug

    options = types.SimpleNamespace()
    for (name, value) in parse_commandline.options_defaults.items():
        setattr(options, name, value)

    progname = dputhelper.get_progname(argv)
    version = dputhelper.get_distribution_version()

    # Parse Command Line Options.
    try:
        (opts, args) = dputhelper.getopt(
                argv[1:],
                "c:dDe:fhHlUopPsuvV", [
                    "debug", "dinstall", "check-only",
                    "check-version", "config=", "force", "help",
                    "host-list", "lintian", "no-upload-log",
                    "passive", "print", "simulate", "unchecked",
                    "delayed=", "version"])
    except dputhelper.DputException as exc:
        sys.stderr.write("{}\n".format(exc))
        sys.exit(os.EX_USAGE)

    # Default to requiring at least one positional argument.
    positional_args_expected = True

    for option, arg in opts:
        if option in ("-h", "--help"):
            sys.stdout.write(make_usage_message())
            sys.exit(os.EX_OK)
        elif option in ("-v", "--version"):
            sys.stdout.write("{progname} {version}\n".format(
                    progname=progname, version=version))
            sys.exit(os.EX_OK)
        elif option in ("-d", "--debug"):
            debug = True
            options.debug = True
        elif option in ("-D", "--dinstall"):
            options.run_dinstall = True
        elif option in ("-c", "--config"):
            options.config_file_path = arg
        elif option in ("-f", "--force"):
            options.force_upload = True
        elif option in ("-H", "--host-list"):
            options.config_host_list = True
            positional_args_expected = False
        elif option in ("-l", "--lintian"):
            options.run_lintian = True
        elif option in ("-U", "--no-upload-log"):
            options.upload_log = False
        elif option in ("-o", "--check-only"):
            options.check_only = True
        elif option in ("-p", "--print"):
            options.config_print = True
            positional_args_expected = False
        elif option in ("-P", "--passive"):
            options.commandline_passive = True
        elif option in ("-s", "--simulate"):
            options.simulate = True
        elif option in ("-u", "--unchecked"):
            options.allow_unsigned_uploads = True
        elif option in ("-e", "--delayed"):
            options.delay_days_text = arg
        elif option in ("-V", "--check-version"):
            options.check_version = True
        # Legacy mis-spelled option name.
        elif option == "--check_version":
            options.check_version = True

    if ((len(args) < 1) and positional_args_expected):
        sys.stdout.write(
                "No package or host has been provided, see dput -h\n")
        sys.exit(os.EX_USAGE)

    return (options, args)


parse_commandline.options_defaults = {
    'debug': debug,
    'check_version': False,
    'config_print': False,
    'force_upload': False,
    'run_lintian': False,
    'config_host_list': False,
    'commandline_passive': None,
    'preferred_host': "",
    'config_file_path': "",
    'run_dinstall': False,
    'upload_log': True,
    'check_only': False,
    'allow_unsigned_uploads': False,
    'delay_days_text': None,
    'simulate': False,
}


def main():
    """ Mainline function for this program. """

    progname = dputhelper.get_progname()
    version = dputhelper.get_distribution_version()

    # Emit the program version in the debug output.
    if debug:
        sys.stdout.write(
                "D: {progname} {version}\n".format(
                    progname=progname, version=version))

    try:
        (options, args) = parse_commandline(sys.argv)
    except SystemExit as exc:
        return exc.code

    # Open and parse all relevant configuration files.
    try:
        config_files = list(configfile.active_config_files(
                options.config_file_path, options.debug))
        config = read_configs(config_files, options.debug)
    except configfile.ConfigurationError as exc:
        sys.stderr.write("Error: {0}\n".format(str(exc)))
        sys.stderr.write("Failed to read program configuration; aborting.\n")
        sys.exit(1)
    if options.config_print:
        print_config(config, options.debug)
        sys.exit(os.EX_OK)

    if options.config_host_list:
        print_default_upload_method(config)
        print_host_list(config)
        sys.exit(os.EX_OK)

    # Process further command line options.
    if len(args) == 1 and not options.check_only:
        changes_file_paths = args[0:]
    else:
        if ':' in args[0]:
            sys.stdout.write("D: Splitting host argument out of  %s.\n" % args[0])
            args[0], host_argument = args[0].split(":", 1)
        else:
            host_argument = ""

        if config.has_section(args[0]):
            sys.stdout.write("D: Setting host argument.\n")
            config.set(args[0], args[0], host_argument)
        else:
            # Let the code below handle this as it is sometimes okay (ie. -o)
            pass

        if not options.check_only:
            if options.debug:
                sys.stdout.write(
                        "D: Checking if a host was named"
                        " on the command line.\n")
            if config.has_section(args[0]):
                if options.debug:
                    sys.stdout.write(
                            "D: Host {} found in config.\n".format(args[0]))
                # Host was also named, so only the rest will be a list
                # of packages to upload.
                options.preferred_host = args[0]
                changes_file_paths = args[1:]
            elif (
                    not config.has_section(args[0])
                    and not args[0].endswith(".changes")):
                sys.stderr.write(
                        "No host {} found in config.\n".format(args[0]))
                sys.exit(1)
            else:
                if options.debug:
                    sys.stdout.write("D: No host named on command line.\n")
                # Only packages have been named on the command line.
                options.preferred_host = ""
                changes_file_paths = args[0:]
        else:
            if options.debug:
                sys.stdout.write("D: Checking for the package name.\n")
            if config.has_section(args[0]):
                sys.stdout.write(
                        "D: Host {} found in config.\n".format(args[0]))
                options.preferred_host = args[0]
                changes_file_paths = args[1:]
            elif len(args) > 1:
                sys.stderr.write("No host %s found in config.\n" % args[0])
                if options.debug:
                    sys.stdout.write("D: No host named on command line.\n")
                changes_file_paths = args[1:]
            else:
                sys.stderr.write(
                        "No host {} found in config.\n".format(args[0]))
                if options.debug:
                    sys.stdout.write("D: No host named on command line.\n")
                changes_file_paths = args[0:]

    upload_methods = import_upload_functions()

    # Run the same checks for all packages that have been given on
    # the command line
    for changes_file_path in changes_file_paths:
        # Check that a .changes file was given on the command line.
        if (not changes_file_path.endswith(".changes")):
            sys.stderr.write(
                    "Filename does not match ‘*.changes’: {}\n".format(
                        changes_file_path))
            sys.exit(1)

        # Construct the package name for further usage.
        changes_file_directory, changes_file_name = os.path.split(
                changes_file_path)
        if changes_file_directory == "":
            changes_file_directory = os.getcwd()

        # Define the host to upload to.
        if options.preferred_host:
            host = options.preferred_host
        else:
            host = guess_upload_host(
                    changes_file_directory, changes_file_name, config)
        fqdn = get_fqdn_for_host(config, host)

        if not config.has_section(host):
            sys.stdout.write("E: No host %s found in config" % host)
            sys.exit(1)

        # Check if we already did this upload or not.
        check_upload_logfile(
                changes_file_path, host, fqdn,
                options.check_only,
                options.force_upload,
                options.debug)

        # Run the change file tests.
        files_to_upload = verify_files(
                changes_file_directory, changes_file_name, host,
                config,
                options.check_only,
                options.check_version,
                options.allow_unsigned_uploads,
                options.debug)

        # Run the lintian test if the user asked us to do so.
        if (
                options.run_lintian or
                config.getboolean(host, 'run_lintian') == 1):
            run_lintian_test(
                    os.path.join(changes_file_directory, changes_file_name))

        if options.check_only:
            # Don't upload, skip to the next item.
            sys.stdout.write("Package checked by dput.\n")
            continue

        if len(config.get(host, 'pre_upload_command')) != 0:
            # Execute commands specified for before each upload.
            position = 'pre'
            command = config.get(host, 'pre_upload_command')
            execute_command(command, position, options.debug)

        verify_config_upload_methods(
                config, host, upload_methods, debug=debug)

        login = get_login_for_host(config, host, debug=debug)

        incoming = get_incoming_for_host(config, host)

        host_config = config[host]
        try:
            delay_days = determine_delay_days(
                    options=options,
                    host_config=host_config,
            )
        except (TypeError, ValueError) as exc:
            sys.stdout.write(
                    "Incorrect delayed argument: {exc_class}: {exc}\n".format(
                        exc_class=type(exc).__name__, exc=exc))
            sys.exit(os.EX_USAGE)
        options.delay_days = delay_days

        # Do the actual upload.
        upload_files(
                upload_methods, config, host, files_to_upload,
                simulate=options.simulate,
                delay_days=options.delay_days,
                commandline_passive=options.commandline_passive,
                debug=options.debug)

        if not options.simulate:
            # Create the log file recording that the package has been sent to
            # the archive.
            if options.upload_log:
                create_upload_file(
                        changes_file_name, host, fqdn, changes_file_directory,
                        files_to_upload, options.debug)
            sys.stdout.write("Successfully uploaded packages.\n")
        else:
            sys.stdout.write("Simulated upload.\n")

        # Run dinstall if the user asked us to do so.
        if options.debug:
            sys.stdout.write(
                    "D: run_dinstall: {}\n".format(options.run_dinstall))
            sys.stdout.write(
                    "D: Host Config: {}\n".format(
                        config.getboolean(host, 'run_dinstall')))
        if (
                config.getboolean(host, 'run_dinstall') == 1
                or options.run_dinstall):
            if not options.simulate:
                dinstall_caller(
                        changes_file_name, host, fqdn, login, incoming,
                        options.debug)
            else:
                sys.stderr.write(
                    "Not running ‘dinstall’: ‘simulate’ option specified.\n")

        if len(config.get(host, 'post_upload_command')) != 0:
            # Execute commands specified for after each upload.
            position = 'post'
            command = config.get(host, 'post_upload_command')
            execute_command(command, position, options.debug)


# Copyright © 2015–2024 Ben Finney <bignose@debian.org>
# Copyright © 2008–2013 Y Giridhar Appaji Nag <appaji@debian.org>
# Copyright © 2006–2008 Thomas Viehmann <tv@beamnet.de>
# Copyright © 2000–2005 Christian Kurz <shorty@debian.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :



# dput/methods/sftp.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# Copyright © 2006-2018 Canonical Ltd.
# Copyright © 2006 Robey Pointer <robey@lag.net>
#                  (parts of ProcessAsChannelAdapter)
# Copyright 2025 Simon Quigley <tsimonq2@ubuntu.com>
#
# Authors: Cody A.W. Somerville <cody.somerville@canonical.com>
#          Julian Andres Klode <julian.klode@canonical.com>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.

import errno
import os
import random
import sys
import socket
import subprocess
import time

from ..helper import dputhelper  # Required for progress tracking

BUFFER_SIZE = 512 * 1024

# Detect if terminal supports colors
USE_COLOR = sys.stdout.isatty()


def colorize(text, color_code):
    """Wrap text with ANSI color codes if terminal supports it."""
    if USE_COLOR:
        return f"\033[{color_code}m{text}\033[0m"
    return text


class ProcessAsChannelAdapter:
    """Wrap a process as a paramiko Channel."""

    def __init__(self, argv):
        """Initialize SSH process and create a socket pair."""
        self.__socket, subproc_sock = socket.socketpair()
        self.__proc = subprocess.Popen(argv, stdin=subproc_sock,
                                       stdout=subproc_sock, bufsize=0)

    def get_name(self):
        """Return a simple name for the adapter."""
        return "ProcessAsChannelAdapter"

    def send(self, data):
        """Send bytes to the subprocess."""
        return self.__socket.send(data)

    def recv(self, num_bytes):
        """Receive bytes from the subprocess."""
        try:
            return self.__socket.recv(num_bytes)
        except socket.error as e:
            if e.args[0] in (errno.EPIPE, errno.ECONNRESET, errno.ECONNABORTED, errno.EBADF):
                return b''  # Paramiko expects an empty string when the connection is closed.
            raise

    def close(self):
        """Close the connection and terminate the process."""
        self.__socket.close()
        self.__proc.terminate()
        self.__proc.wait()


def get_ssh_command_line(login, fqdn, port):
    """Build SSH command line arguments."""
    return [
        "ssh",
        "-oForwardX11=no",
        "-oForwardAgent=no",
        "-oPermitLocalCommand=no",
        "-oClearAllForwardings=yes",
        "-oProtocol=2",
        "-oNoHostAuthenticationForLocalhost=yes",
        "-p", port,
        "-l", login,
        "-s", "--", fqdn, "sftp"
    ]


def human_readable_size(size_bytes):
    """Convert bytes to human-readable format (B, KB, MB, GB, TB)."""
    units = ["B", "KB", "MB", "GB", "TB"]
    size = float(size_bytes)
    for unit in units:
        if size < 1024:
            return f"{size:.2f} {unit}"
        size /= 1024
    return f"{size:.2f} PB"


def copy_file(sftp_client, local_path, remote_path, debug, progress):
    """Upload a single file with detailed progress tracking."""
    try:
        size = os.stat(local_path).st_size
    except Exception:
        size = -1
        if debug:
            sys.stdout.write(f"{colorize('D: Determining size of file failed', '31')}\n")

    file_buffer = min(BUFFER_SIZE, size) if size > 0 else BUFFER_SIZE
    filename = os.path.basename(local_path)

    with open(local_path, 'rb') as fileobj:
        if progress:
            fileobj = dputhelper.FileWithProgress(fileobj, ptype=progress,
                                                  progressf=sys.stdout,
                                                  size=size)

        tmp_path = f"{remote_path}.tmp.{time.time():.9f}.{os.getpid()}.{random.randint(0, 0x7FFFFFFF)}"

        try:
            sys.stdout.write(f"  Uploading {colorize(filename, '36;1')}: {colorize('0.00% (0.00 B / ' + human_readable_size(size) + ')', '32')} ")
            sys.stdout.flush()

            with sftp_client.file(tmp_path, "w", bufsize=0) as remote_fileobj:
                uploaded = 0
                while True:
                    data = fileobj.read(file_buffer)
                    if not data:
                        break
                    remote_fileobj.write(data)
                    uploaded += len(data)

                    percent_complete = (uploaded / size) * 100 if size > 0 else 100
                    progress_str = f"{human_readable_size(uploaded)} / {human_readable_size(size)} ({percent_complete:.2f}%)"
                    sys.stdout.write(f"\r  Uploading {colorize(filename, '36;1')}: {colorize(progress_str, '33')} ")
                    sys.stdout.flush()

            sftp_client.rename(tmp_path, remote_path)

            sys.stdout.write(f"\r  Uploading {colorize(filename, '36;1')}: {colorize(human_readable_size(size) + ' / ' + human_readable_size(size) + ' (100.00%) - done.', '92')}\n")
            sys.stdout.flush()

        except Exception as e:
            sftp_client.remove(tmp_path)
            sys.stderr.write(f"\n{colorize(f'E: Error uploading {filename}: {e}', '31')}\n")
            raise e


def upload(fqdn, login, incoming, files, debug, compress, progress):
    """Upload multiple files via SFTP, with correct progress tracking."""

    try:
        import paramiko.sftp_client
    except ImportError:
        sys.stderr.write(colorize("E: Paramiko must be installed to use SFTP transport.\n", "31"))
        sys.exit(1)

    if ':' in fqdn:
        fqdn, port = fqdn.rsplit(":", 1)
    else:
        port = "22"

    if not login or login == '*':
        login = os.getenv("USER")

    if not incoming.endswith("/"):
        incoming += "/"

    try:
        channel = ProcessAsChannelAdapter(get_ssh_command_line(login, fqdn, port))
        sftp_client = paramiko.sftp_client.SFTPClient(channel)
    except Exception as e:
        sys.stderr.write(f"\n{colorize(f'E: Error connecting to remote host: {e}', '31')}\n")
        sys.exit(1)

    try:
        for local_path in files:
            remote_path = os.path.join(incoming, os.path.basename(local_path))
            try:
                copy_file(sftp_client, local_path, remote_path, debug, progress)
            except Exception as e:
                sys.stderr.write(f"\n{colorize(f'E: Error uploading {os.path.basename(local_path)}: {e}', '31')}\n")
                sys.exit(1)
    finally:
        channel.close()



# dput/configfile.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" Program configuration file functionality. """

import collections
import os.path
import sys

import xdg


class ConfigurationError(RuntimeError):
    """ Exception raised when failing to configure the program. """


default_config_paths = collections.OrderedDict([
        ('global', os.path.join(os.path.sep, "etc", "dput.cf")),
        ('user_xdg', os.path.join(
            xdg.xdg_config_home(), "dput", "dput.cf")),
        ('user_home', os.path.join(
            os.path.expanduser("~"), ".dput.cf")),
])


def active_config_files(config_path, debug):
    """ Yield the sequence of configuration files active for this process.

        :param config_path: Filesystem path of config file to read.
        :param debug: If true, enable debugging output.
        :return: Generator for the sequence of configuration files.
        :raise ConfigurationError: When none of the candidate configuration
            files can be opened.

        The `config_path` specifies a single candidate configuration file path
        to open. If `None`, instead use the default candidate files:

        * The global configuration file ‘/etc/dput.cf’.
        * The user configuration file:
          * $XDG_CONFIG_HOME/dput/dput.cf or, if that does not exist:
          * $HOME/.dput.cf

        Each file is open for reading when yielded in the sequence. When a
        candidate file fails to open (because it does not exist, the process
        does not have read permission to the file, etc.), that file is skipped.
        If all candidate files fail to open, raise `ConfigurationError`.
        """
    def candidate_config_paths(config_path, debug):
        if config_path:
            # Configuration file path explicitly specified to this process.
            yield config_path
        else:
            # Default configuration file paths.
            yield from default_config_paths.values()

    attempted_file_paths = []
    active_file_paths = []
    for infile_path in candidate_config_paths(config_path, debug):
        if not config_path:
            # Iterating through the default configuration paths.
            if (default_config_paths['user_xdg'] in active_file_paths):
                # We already have the standard user configuration file.
                if (infile_path == default_config_paths['user_home']):
                    # No need to try the deprecated location.
                    continue

        attempted_file_paths.append(infile_path)
        try:
            infile = open(infile_path, encoding='UTF-8')
        except OSError as exc:
            if debug:
                sys.stderr.write(
                        "{error}: skipping ‘{path}’\n".format(
                            error=str(exc), path=infile_path))
            continue
        active_file_paths.append(infile_path)
        yield infile

    if not active_file_paths:
        paths_text = ", ".join(
                "‘{0}’".format(path)
                for path in attempted_file_paths)
        raise ConfigurationError(
                "Could not open any configfile, tried {paths}".format(
                    paths=paths_text))


# Copyright © 2016–2024 Ben Finney <bignose@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :



# test/test_configfile.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" Unit tests for config file behaviour. """

import configparser
import contextlib
import doctest
import io
import os
import os.path
import sys
import tempfile
import textwrap
import unittest.mock

import testscenarios
import testtools
import testtools.matchers
import xdg

import dput.configfile
import dput.dput

from .helper import (
        EXIT_STATUS_FAILURE,
        FakeSystemExit,
        FileDouble,
        make_unique_slug,
        patch_os_isatty,
        patch_os_path_exists,
        patch_system_interfaces,
        setup_file_double_behaviour,
        )
from .test_dputhelper import (
        patch_get_username_from_system,
        )


def make_config_from_stream(stream):
    """ Make a ConfigParser parsed configuration from the stream content.

        :param stream: Text stream content of a config file.
        :return: The resulting config if the content parses correctly,
            or ``None``.
        """
    config = configparser.ConfigParser(
            defaults={
                'allow_unsigned_uploads': "false",
                },
            )

    config_file = io.StringIO(stream)
    try:
        config.read_file(config_file)
    except configparser.ParsingError:
        config = None

    return config


def make_config_file_scenarios(config_file_paths=None):
    """ Make a collection of scenarios for testing with config files.

        :return: A collection of scenarios for tests involving config files.

        The collection is a mapping from scenario name to a dictionary of
        scenario attributes.
        """
    if (config_file_paths is None):
        config_file_paths = {}
    runtime_config_file_path = config_file_paths.get(
            'runtime', tempfile.mktemp())
    global_config_file_path = config_file_paths.get(
            'global', os.path.join(os.path.sep, "etc", "dput.cf"))
    user_xdg_config_file_path = config_file_paths.get(
            'user_xdg', os.path.join(xdg.xdg_config_home(), "dput", "dput.cf"))
    user_home_config_file_path = config_file_paths.get(
            'user_home', os.path.join(os.path.expanduser("~"), ".dput.cf"))

    fake_file_empty = io.StringIO()
    fake_file_bogus = io.StringIO("b0gUs")
    fake_file_minimal = io.StringIO(textwrap.dedent("""\
            [DEFAULT]
            """))
    fake_file_comments = io.StringIO(textwrap.dedent("""\
            # Lorem ipsum, dolor sit amet.
            # Lȏrėm îpsüm, dølòr sĭt ámέt.
            [DEFAULT]
            """))
    fake_file_simple = io.StringIO(textwrap.dedent("""\
            [DEFAULT]
            hash = md5
            [foo]
            method = ftp
            fqdn = quux.example.com
            incoming = quux
            check_version = false
            allow_unsigned_uploads = false
            allowed_distributions =
            run_dinstall = false
            """))
    fake_file_simple_host_three = io.StringIO(textwrap.dedent("""\
            [DEFAULT]
            hash = md5
            [foo]
            method = ftp
            fqdn = quux.example.com
            incoming = quux
            check_version = false
            allow_unsigned_uploads = false
            allowed_distributions =
            run_dinstall = false
            [bar]
            fqdn = xyzzy.example.com
            incoming = xyzzy
            [baz]
            fqdn = chmrr.example.com
            incoming = chmrr
            """))
    fake_file_method_local = io.StringIO(textwrap.dedent("""\
            [foo]
            method = local
            incoming = quux
            """))
    fake_file_missing_fqdn = io.StringIO(textwrap.dedent("""\
            [foo]
            method = ftp
            incoming = quux
            """))
    fake_file_missing_incoming = io.StringIO(textwrap.dedent("""\
            [foo]
            method = ftp
            fqdn = quux.example.com
            """))
    fake_file_default_not_unsigned = io.StringIO(textwrap.dedent("""\
            [DEFAULT]
            allow_unsigned_uploads = false
            [foo]
            method = ftp
            fqdn = quux.example.com
            """))
    fake_file_default_distribution_only = io.StringIO(textwrap.dedent("""\
            [DEFAULT]
            default_host_main = consecteur
            [ftp-master]
            method = ftp
            fqdn = quux.example.com
            """))
    fake_file_distribution_none = io.StringIO(textwrap.dedent("""\
            [foo]
            method = ftp
            fqdn = quux.example.com
            distributions =
            """))
    fake_file_distribution_one = io.StringIO(textwrap.dedent("""\
            [foo]
            method = ftp
            fqdn = quux.example.com
            distributions = spam
            """))
    fake_file_distribution_three = io.StringIO(textwrap.dedent("""\
            [foo]
            method = ftp
            fqdn = quux.example.com
            distributions = spam,eggs,beans
            """))

    default_scenario_params = {
            'runtime': {
                'file_double_params': dict(
                    path=runtime_config_file_path,
                    fake_file=fake_file_comments),
                'os_path_exists_scenario_name': 'exist',
                'open_scenario_name': 'okay',
                },
            'global': {
                'file_double_params': dict(
                    path=global_config_file_path,
                    fake_file=fake_file_comments),
                'os_path_exists_scenario_name': 'exist',
                'open_scenario_name': 'okay',
                },
            'user_xdg': {
                'file_double_params': dict(
                    path=user_xdg_config_file_path,
                    fake_file=fake_file_comments),
                'os_path_exists_scenario_name': 'exist',
                'open_scenario_name': 'okay',
                },
            'user_home': {
                'file_double_params': dict(
                    path=user_home_config_file_path,
                    fake_file=fake_file_comments),
                'os_path_exists_scenario_name': 'exist',
                'open_scenario_name': 'okay',
                },
            }

    scenarios = {
            'default': {
                'configs_by_name': {
                    'runtime': None,
                    },
                'expected_config_names': ['global', 'user_xdg'],
                },
            'not-exist': {
                'configs_by_name': {
                    'runtime': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    },
                'expected_config_names': [],
                },
            'exist-read-denied': {
                'configs_by_name': {
                    'runtime': {
                        'open_scenario_name': 'read_denied',
                        },
                    },
                'expected_config_names': [],
                },
            'exist-empty': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_empty),
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-invalid': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_bogus),
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-minimal': {
                'expected_config_names': ['runtime'],
                },
            'exist-simple': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_simple),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-simple-host-three': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_simple_host_three),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-method-local': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_method_local),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-missing-fqdn': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_missing_fqdn),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-missing-incoming': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_missing_incoming),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-default-not-unsigned': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_default_not_unsigned),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-default-distribution-only': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_default_distribution_only),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-distribution-none': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_distribution_none),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-distribution-one': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_distribution_one),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'exist-distribution-three': {
                'configs_by_name': {
                    'runtime': {
                        'file_double_params': dict(
                            path=runtime_config_file_path,
                            fake_file=fake_file_distribution_three),
                        'test_section': "foo",
                        },
                    },
                'expected_config_names': ['runtime'],
                },
            'global-config-not-exist': {
                'configs_by_name': {
                    'global': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['user_xdg'],
                },
            'global-config-read-denied': {
                'configs_by_name': {
                    'global': {
                        'open_scenario_name': 'read_denied',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['user_xdg'],
                },
            'user-xdg-config-not-exist': {
                'configs_by_name': {
                    'user_xdg': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'user_home': {
                        'open_scenario_name': 'okay',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['global', 'user_home'],
                },
            'user-xdg-config-read-denied': {
                'configs_by_name': {
                    'user_xdg': {
                        'open_scenario_name': 'read_denied',
                        },
                    'user_home': {
                        'open_scenario_name': 'okay',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['global', 'user_home'],
                },
            'user-xdg-config-not-exist-user-home-config-not-exist': {
                'configs_by_name': {
                    'user_xdg': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'user_home': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['global'],
                },
            'user-xdg-config-not-exist-user-home-config-read-denied': {
                'configs_by_name': {
                    'user_xdg': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'user_home': {
                        'open_scenario_name': 'read_denied',
                        },
                    'runtime': None,
                    },
                'expected_config_names': ['global'],
                },
            'all-not-exist': {
                'configs_by_name': {
                    'global': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'user_xdg': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'user_home': {
                        'os_path_exists_scenario_name': 'not_exist',
                        'open_scenario_name': 'nonexist',
                        },
                    'runtime': None,
                    },
                'expected_config_names': [],
                },
            'all-read-denied': {
                'configs_by_name': {
                    'global': {
                        'open_scenario_name': 'read_denied',
                        },
                    'user_xdg': {
                        'open_scenario_name': 'read_denied',
                        },
                    'user_home': {
                        'open_scenario_name': 'read_denied',
                        },
                    'runtime': None,
                    },
                'expected_config_names': [],
                },
            }

    for scenario in scenarios.values():
        scenario['empty_file'] = fake_file_empty
        if 'configs_by_name' not in scenario:
            scenario['configs_by_name'] = {}
        for (config_name, default_params) in default_scenario_params.items():
            if config_name not in scenario['configs_by_name']:
                params = default_params
            elif scenario['configs_by_name'][config_name] is None:
                continue
            else:
                params = default_params.copy()
                params.update(scenario['configs_by_name'][config_name])
            params['file_double'] = FileDouble(**params['file_double_params'])
            params['file_double'].set_os_path_exists_scenario(
                    params['os_path_exists_scenario_name'])
            params['file_double'].set_open_scenario(
                    params['open_scenario_name'])
            params['config'] = make_config_from_stream(
                    params['file_double'].fake_file.getvalue())
            scenario['configs_by_name'][config_name] = params

    return scenarios


def get_file_doubles_from_config_file_scenarios(scenarios):
    """ Get the `FileDouble` instances from config file scenarios.

        :param scenarios: Collection of config file scenarios.
        :return: Collection of `FileDouble` instances.
        """
    doubles = set()
    for scenario in scenarios:
        configs_by_name = scenario['configs_by_name']
        doubles.update(
                configs_by_name[config_name]['file_double']
                for config_name in ['global', 'user_home', 'runtime']
                if configs_by_name[config_name] is not None)

    return doubles


def setup_config_file_fixtures(testcase, config_file_paths=None):
    """ Set up fixtures for config file doubles. """
    if (config_file_paths is None):
        config_file_paths = dput.configfile.default_config_paths
    scenarios = make_config_file_scenarios(
            config_file_paths=config_file_paths)
    testcase.config_file_scenarios = scenarios

    setup_file_double_behaviour(
            testcase,
            get_file_doubles_from_config_file_scenarios(scenarios.values()))
    patch_os_path_exists(testcase)


def set_config(testcase, name):
    """ Set the config scenario for a specific test case. """
    scenarios = make_config_file_scenarios()
    testcase.config_scenario = scenarios[name]


def patch_runtime_config_options(testcase):
    """ Patch specific options in the runtime config. """
    config_params_by_name = testcase.config_scenario['configs_by_name']
    runtime_config_params = config_params_by_name['runtime']
    testcase.runtime_config_parser = runtime_config_params['config']

    def maybe_set_option(
            parser, section_name, option_name, value, default=""):
        if value is not None:
            if value is NotImplemented:
                # No specified value. Set a default.
                value = default
            parser.set(section_name, option_name, str(value))
        else:
            # Specifically requested *no* value for the option.
            parser.remove_option(section_name, option_name)

    if testcase.runtime_config_parser is not None:
        testcase.test_host = runtime_config_params.get(
                'test_section', None)

        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'method',
                getattr(testcase, 'config_default_method', "ftp"))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'login',
                getattr(testcase, 'config_default_login', "username"))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'scp_compress',
                getattr(testcase, 'config_default_scp_compress', False))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'ssh_config_options',
                getattr(testcase, 'config_default_ssh_config_options', ""))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'distributions',
                getattr(testcase, 'config_default_distributions', ""))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'incoming',
                getattr(testcase, 'config_default_incoming', "quux"))
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'allow_dcut',
                getattr(testcase, 'config_default_allow_dcut', True))

        config_default_default_host_main = getattr(
                testcase, 'config_default_default_host_main', NotImplemented)
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'default_host_main',
                config_default_default_host_main,
                default="")
        config_default_delayed = getattr(
                testcase, 'config_default_delayed', NotImplemented)
        maybe_set_option(
                testcase.runtime_config_parser,
                'DEFAULT', 'delayed', config_default_delayed,
                default=7)

        for section_name in testcase.runtime_config_parser.sections():
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'method',
                    getattr(testcase, 'config_method', "ftp"))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'login',
                    getattr(testcase, 'config_login', "username"))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'fqdn',
                    getattr(testcase, 'config_fqdn', "quux.example.com"))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'passive_ftp',
                    getattr(testcase, 'config_passive_ftp', False))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'run_lintian',
                    getattr(testcase, 'config_run_lintian', False))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'run_dinstall',
                    getattr(testcase, 'config_run_dinstall', False))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'pre_upload_command',
                    getattr(testcase, 'config_pre_upload_command', ""))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'post_upload_command',
                    getattr(testcase, 'config_post_upload_command', ""))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'progress_indicator',
                    getattr(testcase, 'config_progress_indicator', 0))
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'allow_dcut',
                    getattr(testcase, 'config_allow_dcut', True))
            config_incoming = getattr(
                    testcase, 'config_incoming', NotImplemented)
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'incoming', config_incoming,
                    default=tempfile.mktemp())
            config_delayed = getattr(
                    testcase, 'config_delayed', NotImplemented)
            maybe_set_option(
                    testcase.runtime_config_parser,
                    section_name, 'delayed', config_delayed,
                    default=9)

        for (section_type, options) in (
                getattr(testcase, 'config_extras', {}).items()):
            section_name = {
                    'default': "DEFAULT",
                    'host': testcase.test_host,
                    }[section_type]
            for (option_name, option_value) in options.items():
                maybe_set_option(
                        testcase.runtime_config_parser,
                        section_name, option_name, option_value)


def set_config_file_scenario(testcase, name):
    """ Set the configuration file scenario for `testcase`. """
    testcase.config_file_scenario = testcase.config_file_scenarios[name]
    testcase.configs_by_name = testcase.config_file_scenario['configs_by_name']
    for config_params in testcase.configs_by_name.values():
        if config_params is not None:
            config_params['file_double'].register_for_testcase(testcase)


def get_path_for_runtime_config_file(testcase):
    """ Get the path to specify for runtime config file. """
    path = None
    runtime_config_params = testcase.configs_by_name['runtime']
    if runtime_config_params is not None:
        runtime_config_file_double = runtime_config_params['file_double']
        path = runtime_config_file_double.path
    return path


def patch_default_config_paths(testcase, fake_paths=NotImplemented):
    """ Patch the `default_config_paths` object during `testcase`. """
    if (fake_paths is NotImplemented):
        fake_paths = {
                name: testcase.getUniqueString()
                for name in ['global', 'user_xdg', 'user_home', 'lorem_ipsum']
                }
    patcher = unittest.mock.patch.object(
            dput.configfile, "default_config_paths",
            new=fake_paths)
    patcher.start()
    testcase.addCleanup(patcher.stop)


class active_config_files_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `active_config_files` function. """

    scenarios = [
            ('default', {
                'config_scenario_name': 'default',
                'expected_debug_output': None,
                }),
            ('exist-minimal', {
                'config_scenario_name': 'exist-minimal',
                'expected_debug_output': None,
                }),
            ('not-exist', {
                'config_scenario_name': 'not-exist',
                'expected_debug_output': "skipping",
                }),
            ('exist-read-denied', {
                'config_scenario_name': 'exist-read-denied',
                'expected_debug_output': "skipping",
                }),
            ('global-config-not-exist', {
                'config_scenario_name': 'global-config-not-exist',
                'expected_debug_output': "skipping",
                }),
            ('user-xdg-config-not-exist', {
                'config_scenario_name': 'user-xdg-config-not-exist',
                'expected_debug_output': "skipping",
                }),
            ('user-xdg-config-read-denied', {
                'config_scenario_name': 'user-xdg-config-read-denied',
                'expected_debug_output': "skipping",
                }),
            ('user-xdg-config-not-exist-user-home-config-not-exist', {
                'config_scenario_name':
                    'user-xdg-config-not-exist-user-home-config-not-exist',
                'expected_debug_output': "skipping",
                }),
            ('user-xdg-config-not-exist-user-home-config-read-denied', {
                'config_scenario_name':
                    'user-xdg-config-not-exist-user-home-config-read-denied',
                'expected_debug_output': "skipping",
                }),
            ('all-not-exist', {
                'config_scenario_name': 'all-not-exist',
                'expected_debug_output': "skipping",
                }),
            ('all-read-denied', {
                'config_scenario_name': 'all-read-denied',
                'expected_debug_output': "skipping",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)
        patch_default_config_paths(self)
        setup_config_file_fixtures(self)

        set_config_file_scenario(self, self.config_scenario_name)
        self.expected_config_names = (
                self.config_file_scenario['expected_config_names'])
        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        runtime_config_file_path = get_path_for_runtime_config_file(self)
        self.test_args = dict(
                config_path=runtime_config_file_path,
                debug=False,
                )

    def test_emits_skip_message_when_debug_true(self):
        """ Should emit a “skipping” message when `debug` is true. """
        if (not self.expected_debug_output):
            self.skipTest("no skip message expected in this scenario")
        self.test_args['debug'] = True
        with contextlib.suppress(dput.configfile.ConfigurationError):
            all(dput.configfile.active_config_files(**self.test_args))
        expected_output = self.expected_debug_output
        self.assertIn(expected_output, sys.stderr.getvalue())

    def test_emits_no_skip_message_when_debug_false(self):
        """ Should not emit a “skipping” message when `debug` is false. """
        self.test_args['debug'] = False
        with contextlib.suppress(dput.configfile.ConfigurationError):
            all(dput.configfile.active_config_files(**self.test_args))
        unwanted_output = self.expected_debug_output
        self.assertNotIn(unwanted_output, sys.stderr.getvalue())

    def test_yields_expected_config_files(self):
        """ Should yield a sequence of the expected configuration files. """
        if not self.expected_config_names:
            self.skipTest("Normal return not expected in this scenario")
        expected_files_list = [
                open(
                    self.configs_by_name[config_name]['file_double'].path,
                    encoding='UTF-8')
                for config_name in self.expected_config_names]
        result_list = list(
                dput.configfile.active_config_files(**self.test_args))
        self.assertEqual(expected_files_list, result_list)

    def test_raise_error_if_all_config_files_fail(self):
        """
        Should raise `ConfigurationError` if unable to open any config files.
        """
        if self.expected_config_names:
            self.skipTest("No error expected for this scenario")
        expected_message_regex = (
                "^Could not open any configfile, tried .*")
        with testtools.ExpectedException(
                dput.configfile.ConfigurationError,
                value_re=expected_message_regex):
            all(dput.configfile.active_config_files(**self.test_args))


def patch_active_config_files(testcase):
    """ Patch the `active_config_files` function for `testcase`. """
    if not hasattr(testcase, 'fake_config_files'):
        testcase.fake_config_files = [
                FileDouble(path)
                for path in testcase.getUniqueString() for __ in range(3)]

    def fake_active_config_files(*args, **kwargs):
        for config_file in testcase.fake_config_files:
            yield config_file

    func_patcher = unittest.mock.patch.object(
            dput.configfile, "active_config_files", autospec=True,
            side_effect=fake_active_config_files)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class read_configs_TestCase(testtools.TestCase):
    """ Test cases for `read_config` function. """

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)
        setup_config_file_fixtures(self)

        self.test_configparser = configparser.ConfigParser()
        self.mock_configparser_class = unittest.mock.Mock(
                'ConfigParser',
                return_value=self.test_configparser)

        patcher_class_configparser = unittest.mock.patch.object(
                configparser, "ConfigParser",
                new=self.mock_configparser_class)
        patcher_class_configparser.start()
        self.addCleanup(patcher_class_configparser.stop)

        set_config_file_scenario(self, 'exist-minimal')
        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        runtime_config_file_path = get_path_for_runtime_config_file(self)
        config_files = [
                config_params['file_double'].fake_file
                for config_params in self.configs_by_name.values()]
        self.test_args = dict(
                config_files=config_files,
                debug=False,
                )

    def test_creates_new_parser(self):
        """ Should invoke the `ConfigParser` constructor. """
        dput.dput.read_configs(**self.test_args)
        configparser.ConfigParser.assert_called_with()

    def test_returns_expected_configparser(self):
        """ Should return expected `ConfigParser` instance. """
        result = dput.dput.read_configs(**self.test_args)
        self.assertEqual(self.test_configparser, result)

    def test_sets_default_option_values(self):
        """ Should set values for options, in section 'DEFAULT'. """
        option_names = set([
                'login',
                'method',
                'hash',
                'allow_unsigned_uploads',
                'allow_dcut',
                'distributions',
                'allowed_distributions',
                'run_lintian',
                'run_dinstall',
                'check_version',
                'scp_compress',
                'default_host_main',
                'post_upload_command',
                'pre_upload_command',
                'ssh_config_options',
                'passive_ftp',
                'progress_indicator',
                'delayed',
                ])
        result = dput.dput.read_configs(**self.test_args)
        self.assertTrue(option_names.issubset(set(result.defaults().keys())))

    def test_emits_debug_message_on_parsing_config_file(self):
        """ Should emit a debug message when parsing each config file. """
        self.test_args['debug'] = True
        config_file_double = self.configs_by_name['runtime']['file_double']
        dput.dput.read_configs(**self.test_args)
        expected_output = textwrap.dedent("""\
                D: Parsing configuration file ‘{path}’
                """).format(path=config_file_double.path)
        self.assertIn(expected_output, sys.stdout.getvalue())

    def test_calls_sys_exit_if_config_parsing_error(self):
        """ Should call `sys.exit` if a parsing error occurs. """
        set_config_file_scenario(self, 'exist-invalid')
        self.set_test_args()
        self.test_args['debug'] = True
        with testtools.ExpectedException(FakeSystemExit):
            dput.dput.read_configs(**self.test_args)
        expected_output = textwrap.dedent("""\
                Error parsing configuration file:
                ...
                """)
        self.assertThat(
                sys.stderr.getvalue(),
                testtools.matchers.DocTestMatches(
                    expected_output, flags=doctest.ELLIPSIS))
        sys.exit.assert_called_with(EXIT_STATUS_FAILURE)

    def test_sets_fqdn_option_if_local_method(self):
        """ Should set “fqdn” option for “local” method. """
        set_config_file_scenario(self, 'exist-method-local')
        self.set_test_args()
        result = dput.dput.read_configs(**self.test_args)
        runtime_config_params = self.configs_by_name['runtime']
        test_section = runtime_config_params['test_section']
        self.assertTrue(result.has_option(test_section, "fqdn"))

    def test_exits_with_error_if_missing_fqdn(self):
        """ Should exit with error if config is missing 'fqdn'. """
        set_config_file_scenario(self, 'exist-missing-fqdn')
        self.set_test_args()
        with testtools.ExpectedException(FakeSystemExit):
            dput.dput.read_configs(**self.test_args)
        expected_output = textwrap.dedent("""\
                Config error: {host} must have a fqdn set
                """).format(host="foo")
        self.assertIn(expected_output, sys.stderr.getvalue())
        sys.exit.assert_called_with(EXIT_STATUS_FAILURE)

    def test_exits_with_error_if_missing_incoming(self):
        """ Should exit with error if config is missing 'incoming'. """
        set_config_file_scenario(self, 'exist-missing-incoming')
        self.set_test_args()
        with testtools.ExpectedException(FakeSystemExit):
            dput.dput.read_configs(**self.test_args)
        expected_output = textwrap.dedent("""\
                Config error: {host} must have an incoming directory set
                """).format(host="foo")
        self.assertIn(expected_output, sys.stderr.getvalue())
        sys.exit.assert_called_with(EXIT_STATUS_FAILURE)


class print_config_TestCase(testtools.TestCase):
    """ Test cases for `print_config` function. """

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

    def test_invokes_config_write_to_stdout(self):
        """ Should invoke config's `write` method with `sys.stdout`. """
        test_config = make_config_from_stream("")
        mock_config = unittest.mock.Mock(test_config)
        dput.dput.print_config(mock_config, debug=False)
        mock_config.write.assert_called_with(sys.stdout)


class print_default_upload_method_TestCase(testtools.TestCase):
    """ Test cases for `print_default_upload_method` function. """

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                )

    def test_emits_default_method(self):
        """ Should emit default upload method name to stdout. """
        dput.dput.print_default_upload_method(**self.test_args)
        expected_output = textwrap.dedent("""\

                Default Method: {method}

                """).format(
                    method=self.runtime_config_parser.get(
                        'DEFAULT', 'method'))
        self.assertThat(
                sys.stdout.getvalue(),
                testtools.matchers.DocTestMatches(
                    expected_output, doctest.ELLIPSIS))


def patch_print_default_upload_method(testcase):
    """ Patch `print_default_upload_method` function for the `testcase`. """
    func_patcher = unittest.mock.patch.object(
            dput.dput, "print_default_upload_method", autospec=True,
            return_value=testcase.config_login)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class print_host_list_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `print_host_list` function. """

    scenarios = [
            ('default', {}),
            ('config-distribution-one', {
                'config_scenario_name': "exist-distribution-one",
                }),
            ('config-distribution-three', {
                'config_scenario_name': "exist-distribution-three",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        config_scenario_name = getattr(
                self, 'config_scenario_name',
                'exist-simple')
        set_config(self, config_scenario_name)
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                )

    def test_emits_all_config_sections(self):
        """ Should iteratively emit all config sections to stdout. """
        dput.dput.print_host_list(**self.test_args)
        expected_output_lines = [
                "{section} => {fqdn}  (Upload method: {method}...)".format(
                    section=section_name,
                    fqdn=self.runtime_config_parser.get(section_name, 'fqdn'),
                    method=self.runtime_config_parser.get(
                        section_name, 'method'),
                    )
                for section_name in self.runtime_config_parser.sections()
                ] + ["..."]
        expected_output = "\n".join(expected_output_lines)
        self.assertThat(
                sys.stdout.getvalue(),
                testtools.matchers.DocTestMatches(
                    expected_output, doctest.ELLIPSIS))


def patch_print_host_list(testcase):
    """ Patch the `print_host_list` function for the `testcase`. """
    func_patcher = unittest.mock.patch.object(
            dput.dput, "print_host_list", autospec=True)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_upload_method_name_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_upload_method_name_for_host` function. """

    scenarios = [
            ('config-default-method', {
                'config_default_method': "method-from-config-default",
                'config_method': None,
                'expected_result': "method-from-config-default",
                }),
            ('config-host-method', {
                'config_method': "method-from-config-host",
                'expected_result': "method-from-config-host",
                }),
            ('config-default-and-host-method', {
                'config_default_method': "method-from-config-default",
                'config_method': "method-from-config-host",
                'expected_result': "method-from-config-host",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                )

    def test_returns_expected_result(self):
        """ Should return expected method name for host configuration. """
        result = dput.dput.get_upload_method_name_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_upload_method_name_for_host(testcase):
    """ Patch `get_upload_method_name_for_host` function for `testcase`. """
    if not hasattr(testcase, 'get_upload_method_name_for_host_return_value'):
        method_name = testcase.runtime_config_parser.get(
                testcase.test_host, 'method')
        testcase.get_upload_method_name_for_host_return_value = method_name
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_upload_method_name_for_host", autospec=True,
            return_value=testcase.get_upload_method_name_for_host_return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_login_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_login_for_host` function. """

    scenarios = [
            ('config-default-login', {
                'config_extras': {
                    'default': {
                        'login': "login-from-config-default",
                        },
                    },
                'system_username': "login-from-system",
                'expected_result': "login-from-config-default",
                'expected_output_template':
                    "D: Default login {login} used",
                }),
            ('config-host-login', {
                'config_extras': {
                    'host': {
                        'login': "login-from-config-host",
                        },
                    },
                'system_username': "login-from-system",
                'expected_result': "login-from-config-host",
                'expected_output_template':
                    "D: Login {login} from section {host} used",
                }),
            ('config-default-login sentinel', {
                'config_extras': {
                    'default': {
                        'login': "username",
                        },
                    },
                'system_username': "login-from-system",
                'expected_result': "login-from-system",
                'expected_output_template':
                    "D: Neither host {host} nor default login used.",
                }),
            ('config-host-login sentinel', {
                'config_extras': {
                    'host': {
                        'login': "username",
                        },
                    },
                'system_username': "login-from-system",
                'expected_result': "login-from-system",
                'expected_output_template':
                    "D: Neither host {host} nor default login used.",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(
                self,
                getattr(self, 'config_scenario_name', 'exist-simple'))
        patch_runtime_config_options(self)

        patch_get_username_from_system(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                debug=False,
                )

    def test_emits_debug_message_for_discovered_values(self):
        """ Should emit debug message for values discovered. """
        self.test_args['debug'] = True
        dput.dput.get_login_for_host(**self.test_args)
        expected_output = self.expected_output_template.format(
                host=self.test_host, login=self.expected_result)
        self.assertIn(expected_output, sys.stdout.getvalue())

    def test_returns_expected_result(self):
        """ Should return expected login value for host configuration. """
        result = dput.dput.get_login_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_login_for_host(testcase):
    """ Patch the `get_login_for_host` function for the `testcase`. """
    if not hasattr(testcase, 'config_login'):
        testcase.config_login = make_unique_slug(testcase)
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_login_for_host", autospec=True,
            return_value=testcase.config_login)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_fqdn_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_fqdn_for_host` function. """

    scenarios = [
            ('config-method-http', {
                'config_extras': {
                    'host': {
                        'method': "http",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': "fqdn-from-config-host",
                }),
            ('config-method-http port-443', {
                'config_extras': {
                    'host': {
                        'method': "http",
                        'fqdn': "fqdn-from-config-host:443",
                        },
                    },
                'expected_result': "fqdn-from-config-host",
                }),
            ('config-method-ftp', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': "fqdn-from-config-host",
                }),
            ('config-method-ftp port-21', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host:21",
                        },
                    },
                'expected_result': "fqdn-from-config-host",
                }),
            ('config-method-ftp port-42', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host:42",
                        },
                    },
                'expected_result': "fqdn-from-config-host",
                }),
            ('config-method-local', {
                'config_extras': {
                    'host': {
                        'method': "local",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': "localhost",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                )

    def test_returns_expected_result(self):
        """ Should return expected FQDN value for host configuration. """
        result = dput.dput.get_fqdn_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_fqdn_for_host(testcase):
    """ Patch the `get_fqdn_for_host` function for the `testcase`. """
    if hasattr(testcase, 'get_fqdn_for_host_return_value'):
        return_value = testcase.get_fqdn_for_host_return_value
    elif hasattr(testcase, 'config_fqdn'):
        return_value = testcase.config_fqdn
    else:
        return_value = make_unique_slug(testcase)
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_fqdn_for_host", autospec=True,
            return_value=return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_port_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_port_for_host` function. """

    scenarios = [
            ('config-method-http', {
                'config_extras': {
                    'host': {
                        'method': "http",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': None,
                }),
            ('config-method-http port-443', {
                'config_extras': {
                    'host': {
                        'method': "http",
                        'fqdn': "fqdn-from-config-host:443",
                        },
                    },
                'expected_result': 443,
                }),
            ('config-method-ftp', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': None,
                }),
            ('config-method-ftp port-21', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host:21",
                        },
                    },
                'expected_result': 21,
                }),
            ('config-method-ftp port-42', {
                'config_extras': {
                    'host': {
                        'method': "ftp",
                        'fqdn': "fqdn-from-config-host:42",
                        },
                    },
                'expected_result': 42,
                }),
            ('config-method-local', {
                'config_extras': {
                    'host': {
                        'method': "local",
                        'fqdn': "fqdn-from-config-host",
                        },
                    },
                'expected_result': None,
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                )

    def test_returns_expected_result(self):
        """ Should return expected port value for host configuration. """
        result = dput.dput.get_port_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_port_for_host(testcase):
    """ Patch the `get_port_for_host` function for the `testcase`. """
    if hasattr(testcase, 'get_port_for_host_return_value'):
        return_value = testcase.get_port_for_host_return_value
    else:
        return_value = testcase.getUniqueInteger()
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_port_for_host", autospec=True,
            return_value=return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_delayed_days_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_delayed_days` function. """

    scenarios = [
            ('valid numerals', {
                'option_arg_text': "11",
                'expected_result': 11,
                }),
            ('invalid none', {
                'option_arg_text': None,
                'expected_exception': TypeError,
                }),
            ('invalid empty-string', {
                'option_arg_text': "",
                'expected_exception': TypeError,
                }),
            ('invalid not-text', {
                'option_arg_text': object(),
                'expected_exception': TypeError,
                }),
            ('invalid <0', {
                'option_arg_text': "-1",
                'expected_exception': ValueError,
                }),
            ('invalid >15', {
                'option_arg_text': "16",
                'expected_exception': ValueError,
                }),
            ('invalid malformed', {
                'option_arg_text': "b0gUs",
                'expected_exception': TypeError,
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                days_text=self.option_arg_text,
                )

    def test_returns_expected_result(self):
        """ Should give expected result for inputs. """
        if hasattr(self, 'expected_exception'):
            with testtools.ExpectedException(self.expected_exception):
                dput.dput.get_delayed_days(**self.test_args)
        else:
            result = dput.dput.get_delayed_days(**self.test_args)
            self.assertEqual(self.expected_result, result)


def patch_get_delayed_days(testcase):
    """ Patch the `get_delayed_days` function for the `testcase`. """
    if not hasattr(testcase, 'get_delayed_days_return_value'):
        testcase.get_delayed_days_return_value = None
    func_patcher = unittest.mock.patch.object(
            dput.dput, 'get_delayed_days', autospec=True,
            return_value=testcase.get_delayed_days_return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_incoming_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_incoming_for_host` function. """

    scenarios = [
            ('default-incoming', {
                'config_default_incoming': "incoming-from-config-default",
                'config_incoming': None,
                'expected_result': "incoming-from-config-default",
                }),
            ('host-incoming no-trailing-slash', {
                'config_default_incoming': "incoming-from-config-default",
                'config_incoming': "incoming-from-config",
                'expected_result': "incoming-from-config",
                }),
            ('host-incoming with-trailing-slash', {
                'config_default_incoming': "incoming-from-config-default",
                'config_incoming': "incoming-from-config/",
                'expected_result': "incoming-from-config",
                }),
            ]

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()

        set_config(self, 'exist-simple')
        patch_runtime_config_options(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                )

    def test_returns_expected_result(self):
        """ Should return expected incoming path value. """
        result = dput.dput.get_incoming_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_incoming_for_host(testcase):
    """ Patch the `get_incoming_for_host` function for the `testcase`. """
    if not hasattr(testcase, 'get_incoming_for_host_return_value'):
        testcase.get_incoming_for_host_return_value = tempfile.mktemp()
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_incoming_for_host", autospec=True,
            return_value=testcase.get_incoming_for_host_return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class get_progress_indicator_for_host_TestCase(
        testscenarios.WithScenarios,
        testtools.TestCase):
    """ Test cases for `get_progress_indicator_for_host` function. """

    config_scenarios = [
            ('config-default', {
                'expected_result': 0,
                }),
            ('config-specifies-progress', {
                'config_progress_indicator': 23,
                'expected_result': 23,
                }),
            ]

    isatty_scenarios = [
            ('isatty-true', {
                'os_isatty_return_value': True,
                }),
            ('isatty-false', {
                'os_isatty_return_value': False,
                'progress_indicator_override': 0,
                }),
            ]

    scenarios = testscenarios.multiply_scenarios(
            config_scenarios, isatty_scenarios)

    for (scenario_name, scenario) in scenarios:
        if 'progress_indicator_override' in scenario:
            scenario['expected_result'] = scenario[
                    'progress_indicator_override']

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()

        set_config(
                self,
                getattr(self, 'config_scenario_name', 'exist-simple'))
        patch_runtime_config_options(self)

        patch_os_isatty(self)

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                )

    def test_returns_expected_result(self):
        """ Should return expected progress indicator value. """
        result = dput.dput.get_progress_indicator_for_host(**self.test_args)
        self.assertEqual(self.expected_result, result)


def patch_get_progress_indicator_for_host(testcase):
    """ Patch `get_progress_indicator_for_host` function for `testcase`. """
    if not hasattr(testcase, 'get_progress_indicator_for_host_return_value'):
        testcase.get_progress_indicator_for_host_return_value = 0
    func_patcher = unittest.mock.patch.object(
            dput.dput, "get_progress_indicator_for_host", autospec=True,
            return_value=testcase.get_progress_indicator_for_host_return_value)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


class verify_config_upload_methods_BaseTestCase(
        testtools.TestCase):
    """ Base for test cases for `verify_config_upload_methods` function. """

    def setUp(self):
        """ Set up test fixtures. """
        super().setUp()
        patch_system_interfaces(self)

        set_config(
                self,
                getattr(self, 'config_scenario_name', 'exist-simple'))
        patch_runtime_config_options(self)

        self.upload_methods = {
                name: unittest.mock.MagicMock()
                for name in ["local", "ftp", "scp"]}

        self.set_test_args()

    def set_test_args(self):
        """ Set the arguments for the test call to the function. """
        self.test_args = dict(
                config=self.runtime_config_parser,
                host=self.test_host,
                upload_methods=self.upload_methods,
                debug=False,
                )


class verify_config_upload_methods_SuccessTestCase(
        verify_config_upload_methods_BaseTestCase):
    """ Test cases for `verify_config_upload_methods`, when successful. """

    def test_emits_debug_message_for_discovered_methods(self):
        """ Should emit debug message for discovered upload methods. """
        self.test_args['debug'] = True
        dput.dput.verify_config_upload_methods(**self.test_args)
        expected_output = textwrap.dedent("""\
                D: Default Method: {default_method}
                D: Host Method: {host_method}
                """).format(
                    default_method=self.runtime_config_parser.get(
                        'DEFAULT', 'method'),
                    host_method=self.runtime_config_parser.get(
                        self.test_host, 'method'))
        self.assertIn(expected_output, sys.stdout.getvalue())


class verify_config_upload_methods_UnknownUploadMethodTestCase(
        testscenarios.WithScenarios,
        verify_config_upload_methods_BaseTestCase):
    """ Test cases for `verify_config_upload_methods`, when unknown method. """

    scenarios = [
            ('bogus-default-method', {
                'config_extras': {
                    'default': {
                        'method': "b0gUs",
                        },
                    },
                'expected_output': "Unknown upload method: b0gUs",
                }),
            ('bogus-host-method', {
                'config_extras': {
                    'host': {
                        'method': "b0gUs",
                        },
                    },
                'expected_output': "Unknown upload method: b0gUs",
                }),
            ]

    def test_emits_error_message_when_unknown_method(self):
        """ Should emit error message when unknown upload method. """
        with contextlib.suppress(FakeSystemExit):
            dput.dput.verify_config_upload_methods(**self.test_args)
        self.assertIn(self.expected_output, sys.stdout.getvalue())

    def test_calls_sys_exit_when_unknown_method(self):
        """ Should call `sys.exit` when unknown upload method. """
        with testtools.ExpectedException(FakeSystemExit):
            dput.dput.verify_config_upload_methods(**self.test_args)
        sys.exit.assert_called_with(EXIT_STATUS_FAILURE)


def patch_verify_config_upload_methods(testcase):
    """ Patch `verify_config_upload_methods` function for the `testcase`. """
    func_patcher = unittest.mock.patch.object(
            dput.dput, "verify_config_upload_methods", autospec=True)
    func_patcher.start()
    testcase.addCleanup(func_patcher.stop)


# Copyright © 2015–2024 Ben Finney <bignose@debian.org>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :



# dput/methods/rsync.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" Implementation for RSync upload method. """

import os
import sys

from ..helper import dputhelper


rsync_options = [
        "--copy-links", "--progress", "--partial",
        "-zave", "ssh -x",
        ]


def upload(fqdn, login, incoming, files_to_upload, debug, dummy, progress=0):
    """ Upload the files with rsync via ssh. """

    remote_file_paths = []

    for file_path in files_to_upload:
        file_name = os.path.basename(file_path)
        remote_file_path = os.path.join(incoming, file_name)
        remote_file_paths.append(remote_file_path)

    if login and login != "*":
        login_spec = "{}@".format(login)
    else:
        login_spec = ""
    destination = "{login_spec}{fqdn}:{incoming}".format(**vars())
    upload_command = ["rsync", *files_to_upload, *rsync_options, destination]
    fix_command = [
            "ssh", "{login_spec}{fqdn}".format(**vars()), "chmod", "0644"
            ] + remote_file_paths

    if debug:
        sys.stdout.write((
                "D: Uploading with rsync to {login_spec}{fqdn}:{incoming}\n"
                ).format(**vars()))
    if dputhelper.check_call(upload_command) != dputhelper.EXIT_STATUS_SUCCESS:
        sys.stdout.write(
                "\n"
                "Error while uploading.\n")
        sys.exit(1)
    if debug:
        sys.stdout.write((
                "D: Fixing file permissions with {login_spec}{fqdn}\n"
                ).format(**vars()))
    if dputhelper.check_call(fix_command) != dputhelper.EXIT_STATUS_SUCCESS:
        sys.stdout.write("Error while fixing permission.\n")
        sys.exit(1)


# Copyright © 2015–2024 Ben Finney <bignose@debian.org>
# Copyright © 2009 Y Giridhar Appaji Nag <appaji@debian.org>
# Copyright © 2007 Thomas Viehmann <tv@beamnet.de>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :



# dput/methods/http.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" Implementation for HTTP upload method. """

import contextlib
import getpass
from http.client import HTTPConnection
import os
import sys
import urllib.parse
import urllib.request

from ..helper import dputhelper


MESSAGE_BODY_ENCODING = 'utf-8'


class PromptingPasswordMgr(urllib.request.HTTPPasswordMgr):
    """ Password manager that prompts at the terminal.

        Custom HTTP password manager that prompts for a password using
        getpass() if required, and mangles the saved URL so that only
        one password is prompted for.
        """

    def __init__(self, username):
        urllib.request.HTTPPasswordMgr.__init__(self)
        self.username = username

    def find_user_password(self, realm, authuri):
        # Hack so that we only prompt for a password once
        authuri = self.reduce_uri(authuri)[0]
        authinfo = urllib.request.HTTPPasswordMgr.find_user_password(
                self, realm, authuri)
        if authinfo != (None, None):
            return authinfo

        password = getpass.getpass("    Password for {}:".format(realm))
        self.add_password(realm, authuri, self.username, password)
        return (self.username, password)


class AuthHandlerHackAround:
    """ Fake request and parent object. """

    def __init__(self, url, response_headers, password_mgr):
        self.headers = {}
        self.url = url
        self.response_headers = response_headers
        self.authhandlers = []
        self.timeout = {}

        # The API for a `urllib.Request` presents this property.
        self.full_url = url

        # digest untested
        for authhandler_class in [
                urllib.request.HTTPBasicAuthHandler,
                urllib.request.HTTPDigestAuthHandler]:
            ah = authhandler_class(password_mgr)
            ah.add_parent(self)
            self.authhandlers.append(ah)

    def get_header(self, header_name, default=None):
        value = self.headers.get(header_name, default)
        return value

    def add_header(self, key, val):
        self.headers[key] = val

    def add_unredirected_header(self, key, header):
        self.headers[key] = header

    def get_full_url(self):
        return self.full_url

    def open(self, *args, **keywords):
        pass

    def get_auth_headers(self):
        for ah in self.authhandlers:
            with contextlib.suppress(ValueError):
                ah.http_error_401(self, None, 401, None, self.response_headers)
        if self.headers:
            return self.headers
        return self.headers


def upload(
        fqdn, login, incoming, files_to_upload, debug, dummy,
        progress=0, protocol="http", connection_class=HTTPConnection):
    """ Upload the files via WebDAV. """

    if protocol not in ["http", "https"]:
        sys.stderr.write("Wrong protocol for upload http[s].py method\n")
        sys.exit(1)
    incoming = "/{}".format(incoming.strip("/"))
    unprocessed_files_to_upload = files_to_upload[:]
    auth_headers = {}
    password_mgr = PromptingPasswordMgr(login)
    while unprocessed_files_to_upload:
        file_path = unprocessed_files_to_upload[0]
        file_directory, file_name = os.path.split(file_path)
        sys.stdout.write("  Uploading {}: ".format(file_name))
        sys.stdout.flush()
        try:
            size = os.stat(file_path).st_size
        except Exception:
            sys.stderr.write(
                    "Determining size of file '{}' failed\n".format(file_path))
            sys.exit(1)
        f = open(file_path, 'rb')
        if progress:
            f = dputhelper.FileWithProgress(
                    f, ptype=progress,
                    progressf=sys.stderr,
                    size=size)
        url_path = "/".join([incoming, file_name])
        url_fields = urllib.parse.ParseResult(
                scheme=protocol,
                netloc=fqdn,
                path=url_path,
                params=None,
                query=None,
                fragment=None)
        url = urllib.parse.urlunparse(url_fields)
        if debug:
            sys.stdout.write("D: HTTP-PUT to URL: {}\n".format(url))
        conn = connection_class(fqdn)
        conn.putrequest("PUT", url_path, skip_accept_encoding=True)
        # Host: should be automatic
        conn.putheader("User-Agent", "dput")
        for k, v in auth_headers.items():
            conn.putheader(k, v)
        conn.putheader("Connection", "close")
        conn.putheader("Content-Length", str(size))
        conn.endheaders()
        pos = 0
        while pos < size:
            # sending in 64k steps (screws progress a bit)
            s = f.read(65536)
            conn.send(s)
            pos += len(s)
        f.close()
        s = ""
        res = conn.getresponse()
        if res.status >= 200 and res.status < 300:
            sys.stdout.write("done.\n")
            del unprocessed_files_to_upload[0]
        elif res.status == 401 and not auth_headers:
            sys.stdout.write("need authentication.\n")
            auth_headers = AuthHandlerHackAround(
                    url, res.msg, password_mgr).get_auth_headers()
        else:
            if res.status == 401:
                sys.stdout.write(
                        "Upload failed as unauthorized: {0.reason}\n"
                        "  Maybe wrong username or password?\n".format(res))
            else:
                sys.stdout.write(
                        "Upload failed: {0.status:d} {0.reason}\n".format(res))
            response_body = res.read().decode(MESSAGE_BODY_ENCODING)
            if response_body:
                if debug:
                    sys.stdout.write("D: Response body: {body}\n".format(
                            body=response_body))
            sys.exit(1)
        # must be done, but we're not interested
        res.read()


# Copyright © 2015–2024 Ben Finney <bignose@debian.org>
# Copyright © 2008–2012 Y Giridhar Appaji Nag <appaji@debian.org>
# Copyright © 2007–2008 Thomas Viehmann <tv@beamnet.de>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :



# dput/helper/dputhelper.py
# Part of ‘dput’, a Debian package upload toolkit.
#
# This is free software, and you are welcome to redistribute it under
# certain conditions; see the end of this file for copyright
# information, grant of license, and disclaimer of warranty.

""" Helper code for Dput. """

import importlib.metadata
import io
import locale
import os
import pwd
import subprocess
import sys


class DputException(Exception):
    pass


class DputUploadFatalException(DputException):
    pass


class UploadChangesInvalidVersionDputError(DputException):
    """
    Raised when an invalid ‘version’ field from upload changes control data.
    """


EXIT_STATUS_SUCCESS = 0
EXIT_STATUS_FAILURE = 1
EXIT_STATUS_COMMAND_NOT_FOUND = 127


# This wrapper is intended as a migration target for the prior
# `spawnv` wrapper, and will produce the same output for now.
def check_call(args, *posargs, **kwargs):
    """ Wrap `subprocess.check_call` with error output. """
    command_file_path = args[0]
    try:
        subprocess.check_call(args, *posargs, **kwargs)
        exit_status = EXIT_STATUS_SUCCESS
    except subprocess.CalledProcessError as exc:
        exit_status = exc.returncode
        if exit_status == EXIT_STATUS_COMMAND_NOT_FOUND:
            sys.stderr.write(
                    "Error: Failed to execute '{path}'.\n"
                    "       "
                    "The file may not exist or not be executable.\n".format(
                        path=command_file_path))
        else:
            sys.stderr.write(
                    "Warning: The execution of '{path}' as\n"
                    "  '{command}'\n"
                    "  returned a nonzero exit code.\n".format(
                        path=command_file_path, command=" ".join(args)))
    return exit_status


class FileWithProgress:
    """ Mimics a file (passed as f, an open file), but with progress.

        FileWithProgress(f, args)

        args:
        * ptype = 1,2 is the type ("|/-\" or numeric), default 0 (no progress)
        * progressf = file to output progress to (default sys.stdout)
        * size = size of file (or -1, the default, to ignore)
                 for numeric output
        * step = stepsize (default 1024)
        """

    def __init__(self, f, ptype=0, progressf=sys.stdout, size=-1, step=1024):
        self.f = f
        self.count = 0
        self.lastupdate = 0
        self.ptype = ptype
        self.ppos = 0
        self.progresschars = ["|", "/", "-", "\\"]
        self.progressf = progressf
        self.size = size
        self.step = step
        self.closed = 0

    def __getattr__(self, name):
        return getattr(self.f, name)

    def read(self, size=-1):
        a = self.f.read(size)
        self.count = self.count + len(a)
        if (self.count - self.lastupdate) > 1024:
            if self.ptype == 1:
                self.ppos = (self.ppos + 1) % len(self.progresschars)
                self.progressf.write("{backspace}{spinner}".format(
                        backspace=("\b" if self.lastupdate else ""),
                        spinner=self.progresschars[self.ppos]))
                self.progressf.flush()
                self.lastupdate = self.count
            elif self.ptype == 2:
                s = "{size:d}k".format(size=(self.count // self.step))
                if self.size >= 0:
                    s = "{text}/{size:d}k".format(
                            text=s,
                            size=((self.size + self.step - 1) // self.step))
                s = "{text}{spaces}".format(
                        text=s,
                        spaces=(min(self.ppos - len(s), 0) * " "))
                self.progressf.write("{backspaces}{text}".format(
                        backspaces=(self.ppos * "\b"),
                        text=s))
                self.progressf.flush()
                self.ppos = len(s)
        return a

    def close(self):
        if not self.closed:
            self.f.close()
            self.closed = 1
            if self.ptype == 1:
                if self.lastupdate:
                    self.progressf.write("\b \b")
                    self.progressf.flush()
            elif self.ptype == 2:
                self.progressf.write(
                        "{backspaces}{spaces}{backspaces}".format(
                            backspaces=(self.ppos * "\b"),
                            spaces=(self.ppos * " ")))
                self.progressf.flush()

    def __del__(self):
        self.close()


def make_text_stream(stream):
    """ Make a text stream from the specified stream.

        :param stream: An open file-like object.
        :return: A stream object providing text I/O.

        In the normal case, the specified stream is a stream providing
        bytes I/O. We create an `io.TextIOWrapper` with the
        appropriate encoding for the byte stream, and return that
        wrapper stream.

        The text encoding is determined by interrogating the file
        object. If the file object has no declared encoding, the
        default `locale.getpreferredencoding(False)` is used.

        If the stream is a `io.TextIOBase` instance, it is already
        providing text I/O. In this case, the stream is returned as
        is.
        """
    result = None

    if hasattr(stream, 'encoding'):
        encoding = stream.encoding
    else:
        encoding = locale.getpreferredencoding(False)

    if isinstance(stream, io.TextIOBase):
        result = stream
    else:
        result = io.TextIOWrapper(stream, encoding=encoding)

    return result


def get_progname(argv=None):
    """ Get the program name from the command line arguments.

        :param argv: Sequence of command-line arguments.
            Defaults to `sys.argv`.
        :return: The program name used to invoke this program.
        """
    if argv is None:
        argv = sys.argv
    progname = os.path.basename(argv[0])
    return progname


def get_distribution_version():
    """ Get the version string for this distribution. """
    distribution = importlib.metadata.distribution("dput")
    return distribution.version


def getopt(args, shortopts, longopts):
    args = args[:]
    optlist = []
    while args and args[0].startswith("-"):
        if args[0] == "--":
            args = args[1:]
            break
        if args[0] == "-":
            break
        if args[0].startswith("--"):
            opt = args.pop(0)[2:]
            if "=" in opt:
                opt, optarg = opt.split("=", 1)
            else:
                optarg = None
            prefixmatch = [x for x in longopts if x.startswith(opt)]
            if len(prefixmatch) == 0:
                raise DputException("unknown option --{}".format(opt))
            elif len(prefixmatch) > 1:
                raise DputException("non-unique prefix --{}".format(opt))
            opt = prefixmatch[0]
            if opt.endswith("=="):
                opt = opt[:-2]
                optarg = optarg or ""
            elif opt.endswith("="):
                opt = opt[:-1]
                if not optarg:
                    if not args:
                        raise DputException(
                                "option --{} requires argument".format(opt))
                    optarg = args.pop(0)
            else:
                if optarg is not None:
                    raise DputException(
                            "option --{} does not take arguments".format(opt))
                optarg = ""
            optlist.append(("--{}".format(opt), optarg))
        else:
            s = args.pop(0)[1:]
            while s:
                pos = shortopts.find(s[0])
                if pos == -1:
                    raise DputException("option -{} unknown".format(s[0]))
                if pos + 1 >= len(shortopts) or shortopts[pos + 1] != ":":
                    optlist.append(("-{}".format(s[0]), ""))
                    s = s[1:]
                elif len(s) > 1:
                    optlist.append(("-{}".format(s[0]), s[1:]))
                    s = ""
                elif args:
                    optlist.append(("-{}".format(s), args.pop(0)))
                    s = ""
                else:
                    raise DputException(
                            "option -{} requires argument".format(s))
    return optlist, args


def make_person_text_from_fields(*, raw_text, full_name=""):
    """ Make a text representation of a person's `full_name` and `email_addr`.

        :param raw_text: The existing raw text of the field.
        :param full_name: The full name of the person.
        :return: The normalised text representation of the field values.
        """
    if ("<" in raw_text):
        # The raw text already looks like a person text representation.
        person_text = raw_text
    else:
        fields = (
            value for value in [full_name, "<{}>".format(raw_text)]
            if value)
        person_text = " ".join(fields)
    return person_text


def get_username_from_system(debug=False):
    """ Get the system username for the current process.

        :param debug: If true, enable debugging output.
        :return: The username determined from the current process.
        """
    login = None
    if 'USER' in os.environ:
        login = os.environ['USER']
    else:
        sys.stdout.write("$USER not set, will use login information.\n")
        login = pwd.getpwuid(os.getuid())[0]
        if debug:
            sys.stderr.write("D: User-ID: {:d}\n".format(os.getuid()))
    if debug:
        sys.stderr.write("D: Login: {}\n".format(login))

    return login


# Copyright © 2015–2024 Ben Finney <bignose@debian.org>
# Copyright © 2009–2010 Y Giridhar Appaji Nag <appaji@debian.org>
# Copyright © 2007–2008 Thomas Viehmann <tv@beamnet.de>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.


# Local variables:
# coding: utf-8
# mode: python
# End:
# vim: fileencoding=utf-8 filetype=python :
